from .base import *

# from .default_localizer import DefaultLocalizer
from .drift_corrector import CorrelationDriftCorrector, BeadDriftCorrector
from .trackpy_tracker import ConservativeTracker
from .window_percentile import WindowPercentileFilter
from .quot_localizer import (
    MaxLikelihoodLocalizer,
    BaseDetector,
    RadialLocalizer,
)
from .quot_tracker import EuclideanTracker, DiffusionTracker



#!/usr/bin/env python
"""
core.py -- high-level user functions for running 
filtering, detection, subpixel localization, and 
tracking sequentially on the same datasets

"""
# File paths
import os 
from glob import glob 

# Progress bar
from tqdm import tqdm 

# Dataframes
import pandas as pd 

# Parallelization
import dask 
from dask.diagnostics import ProgressBar 

# File readers and filterers
from .chunkFilter import ChunkFilter

# Core detection function
from .findSpots import detect 

# Core localization function
from .subpixel import localize_frame

# Core tracking function
from .track import track 

def localize_file(path, out_csv=None, progress_bar=True, **kwargs):
    """
    Run filtering, detection, and subpixel localization on 
    a single image movie. This does NOT perform tracking.

    args
    ----
        path        :   str, path to the image file
        out_csv     :   str, path to save file, if 
                        desired
        progress_bar:   bool, show a progress bar
        kwargs      :   configuration

    returns
    -------
        pandas.DataFrame, the localizations

    """
    # Make sure the file exists
    assert os.path.isfile(path), "quot.__main__.localize_file: " \
        "file %s does not exist" % path 

    # If the config file does not contain a "filter" section,
    # don't worry about it
    kwargs["filter"] = kwargs.get("filter", {})

    # Open an image file reader with some filtering
    # settings, if desired
    with ChunkFilter(path, **kwargs['filter']) as f:

        frames = enumerate(f)
        if progress_bar:
            frames = tqdm(frames)

        locs = []
        for frame_idx, frame in frames:

            # Find spots in this image frame
            detections = detect(frame, **kwargs['detect'])

            # Localize spots to subpixel resolution
            locs.append(localize_frame(frame, detections, 
                **kwargs['localize']).assign(frame=frame_idx))

        locs = pd.concat(locs, ignore_index=True, sort=False)

        # Adjust for start index
        locs['frame'] += kwargs['filter'].get('start', 0)

    # Save to a file, if desired
    if not out_csv is None:
        locs.to_csv(out_csv, index=False)

    return locs 

def track_file(path, out_csv=None, progress_bar=True, **kwargs):
    """
    Run filtering, detection, subpixel localization, and 
    tracking on a single target movie.

    args
    ----
        path        :   str, path to the image file
        out_csv     :   str, path to save file, if 
                        desired
        progress_bar:   bool, show a progress bar
        kwargs      :   tracking configuration

    returns
    -------
        pandas.DataFrame, the reconnected localizations

    """ 
    # Run filtering + detection + localization
    locs = localize_file(path, out_csv=None, progress_bar=progress_bar,
        **kwargs)

    # Track localizations between frames
    locs = track(locs, **kwargs['track'])

    # Save to a file if desired 
    if not out_csv is None:
        locs.to_csv(out_csv, index=False)

    return locs 

def track_files(paths, num_workers=4, save=True, out_dir=None, **kwargs):
    """
    Run tracking on several files using parallelization.

    args
    ----
        paths       :   list of str, paths to image files to track
        num_workers :   int, the number of threads to use
        save        :   bool, save the output to CSVs files. The names
                        for these CSVs are generated from the names of 
                        the corresponding image files.
        out_dir     :   str, output directory
        kwargs      :   tracking configuration, as read with 
                        quot.io.read_config

    returns
    -------
        list of pandas.DataFrame, the tracking results for each 
            file

    """
    # Create the output directory if it does not already exist
    if (not out_dir is None) and (not os.path.isdir(out_dir)):
        os.mkdir(out_dir)

    # Tracking function for one file with lazy evaluation
    @dask.delayed 
    def driver(path):
        if save and (not out_dir is None):
            out_csv = os.path.join(
                out_dir,
                "{}_trajs.csv".format(os.path.splitext(os.path.basename(path))[0])
            )
        elif save and (out_dir is None):
            out_csv = "{}_trajs.csv".format(os.path.splitext(path)[0])
        else:
            out_csv = None 
        try:
            return track_file(
                path,
                out_csv=out_csv,
                progress_bar=False,
                **kwargs
            )
        except Exception as e:
            print("WARNING: Failed to analyze file {} due to exception:".format(path))
            print(e)
            return []

    # Run localization and tracking on each file
    results = [driver(path) for path in paths]
    scheduler = "single-threaded" if num_workers == 1 else "processes"
    with ProgressBar():
        dask.compute(*results, scheduler=scheduler, num_workers=num_workers)

    return results 

def track_directory(path, ext='.nd2', num_workers=4, save=True, contains=None,
    out_dir=None, **kwargs):
    """
    Find all image files in a directory and run 
    localization and tracking.

    args
    ----
        path        :   str, path to directory
        ext         :   str, image file extension
        num_workers :   int, the number of threads
                        to use 
        save        :   bool, save the results to CSV
                        files
        contains    :   str, a substring that all image files 
                        are required to contain
        out_dir     :   str, directory for output CSV files
        kwargs      :   configuration

    returns
    -------
        None. Output of trackng is saved to files 
            with extension "_trajs.csv" in the same
            directory.

    """
    # Make sure the directory exists
    assert os.path.isdir(path), "quot.__main__.localize_directory: " \
        "directory %s does not exist" % path 

    # Find all image files in this directory
    image_paths = glob("%s/*%s" % (path, ext))

    # Only include image files with a substring, if specified
    if not contains is None:
        image_paths = [j for j in image_paths if contains in os.path.basename(j)]

    # Run tracking and localization
    track_files(image_paths, num_workers=num_workers, save=save, 
        out_dir=out_dir, **kwargs)

def retrack_file(path, out_csv=None, **kwargs):
    """
    Given an existing set of localizations or trajectories, (re)run tracking
    to reconstruct trajectories.

    args
    ----
        path        :   str, path to a *trajs.csv file
        out_csv     :   str, path to save the resulting trajectories, if 
                        desired
        kwargs      :   tracking configuration

    returns
    -------
        pandas.DataFrame, the reconnected localizations 

    """
    # Load the file
    T = pd.read_csv(path)

    # Track localizations between frames
    T = track(T, **kwargs)

    # Save to a file, if desired
    if not out_csv is None:
        T.to_csv(out_csv, index=False)

    return T 

def retrack_files(paths, out_suffix=None, num_workers=1, **kwargs):
    """
    Given a set of localizations, run retracking on each file and save to a 
    CSV.

    If *out_suffix* is not specified, then the trajectories are saved to the 
    original set of localization files (overwriting them).

    args
    ----
        paths       :   list of str, a set of CSV files encoding trajectories
        out_suffix  :   str, the suffix to use when generating the output 
                        paths. If *None*, then the output trajectories are 
                        saved to the original file path.
        num_workers :   int, the number of threads to use
        kwargs      :   tracking configuration

    """
    # Avoid redundant extensions
    if (not out_suffix is None) and (not ".csv" in out_suffix):
        out_suffix = "{}.csv".format(out_suffix)

    @dask.delayed 
    def task(fn):
        """
        Retrack one file.

        """
        out_csv = fn if out_suffix is None else \
            "{}_{}".format(os.path.splitext(fn)[0], out_suffix)
        retrack_file(fn, out_csv=out_csv, **kwargs["track"])

    # Run retracking on all files
    scheduler = "single-threaded" if num_workers == 1 else "processes"
    tasks = [task(fn) for fn in paths]
    with ProgressBar():
        dask.compute(*tasks, num_workers=num_workers, scheduler=scheduler)



#!/usr/bin/env python
"""
trajUtils.py -- functions to compute some common values on 
trajectories

"""
# Numeric
import numpy as np 

# Uniform filter
from scipy.ndimage import uniform_filter 

# DataFrames
import pandas as pd 


#########################################
## OPERATIONS ON TRAJECTORY DATAFRAMES ##
#########################################

def traj_length(trajs):
    """
    Compute the number of localizations corresponding to 
    each trajectory.

    Unassigned localizations (with a trajectory index of 
    -1) are given traj_len = 0.

    args
    ----
        trajs       :   pandas.DataFrame, localizations with the 
                        'trajectory' column

    returns
    -------
        pandas.DataFrame, the same dataframe with "traj_len"
            column

    """
    # Remove traj_len column if it already exists
    if "traj_len" in trajs.columns:
        trajs = trajs.drop("traj_len", axis=1)

    # Get the lengths 
    trajs = trajs.join(
        trajs.groupby("trajectory").size().rename("traj_len"),
        on="trajectory"
    )

    # Unassigned localizations are given a traj_len of 0
    trajs.loc[trajs['trajectory']==-1, 'traj_len'] = 0
    return trajs  

def radial_disps(trajs, pixel_size_um=0.16, first_only=False):
    """
    Calculate the 2D radial displacement of each jump in every 
    trajectory.

    args
    ----
        trajs           :   pandas.DataFrame
        pixel_size_um   :   float, size of pixels
        first_only      :   float, only compute the first 
                            displacement of each trajectory

    returns
    -------
        pandas.DataFrame, with the new column 'radial_disp_um'

    """
    # If handed an empty dataframe, assign a dead column
    if len(trajs) == 0:
        trajs['radial_disp_um'] = []

    # Order by ascending trajectory, frame
    trajs = trajs.sort_values(by=['trajectory', 'frame'])

    # Filter out singlets
    if not "traj_len" in trajs.columns:
        trajs = traj_length(trajs)
    F0 = np.logical_and(trajs['trajectory']>=0, trajs['traj_len']>1)

    # If no locs remain, return all NaN
    if F0.sum() == 0:
        trajs['radial_disp_um'] = np.nan 

    # Format as ndarray for speed
    S = np.asarray(trajs.loc[F0, ['trajectory', 'frame', 'y', 'x']])

    # Compute Cartesian displacements
    D = S[1:,:] - S[:-1,:]

    # Add extra row at the end for indexing purposes (there is one
    # fewer row in D than in S)
    D = np.concatenate((D, np.array([[1, 1, 1, 1]])), axis=0)

    # Take all displacements that originate from the same trajectory
    # and from a single frame interval
    F1 = np.logical_and(D[:,0]==0, D[:,1]==1)

    # If no disps remain, return all NaN
    if F1.sum() == 0:
        trajs['radial_disp_um'] = np.nan 

    # Compute radial displacements
    R = np.full(F0.sum(), np.nan)
    R[F1] = np.sqrt((D[F1, 2:]**2).sum(axis=1))

    # Assign to the corresponding rows in the original dataframe
    trajs.loc[F0, "radial_disp_um"] = R 

    # If desired, only used the first jump from each trajectory
    if first_only:
        trajs["_ones"] = 1 
        trajs["_first_loc"] = trajs.groupby("trajectory")["_ones"].cumsum()==1 
        trajs.loc[~trajs['_first_loc'], 'radial_disp_um'] = np.nan 
        trajs = trajs.drop(['_ones', '_first_loc'], axis=1)

    # Scale into um
    trajs['radial_disp_um'] = trajs['radial_disp_um'] * pixel_size_um 

    return trajs 

def radial_disp_histograms(trajs, n_intervals=1, pixel_size_um=0.16,
    first_only=False, n_gaps=0, bin_size=0.001, max_disp=5.0):
    """
    Calculate radial displacement histograms for a set of 
    trajectories.

    args
    ----
        trajs           :   pandas.DataFrame
        n_intervals     :   int, the maximum number of frames
                            over which to compute displacements
        pixel_size_um   :   float, size of pixels (um)
        first_only      :   bool, only take the displacements
                            relative to the first point of 
                            each trajectory
        n_gaps          :   int, the number of gaps allowed
                            in tracking
        bin_size        :   float, the spatial bin size in um
        max_disp        :   float, maximum displacement to 
                            consider

    returns
    -------
        (
            2D ndarray of shape (n_intervals, n_bins), the radial
                displacements in each bin for each interval;
            1D ndarray of shape (n_bins+1), the edges of the
                displacement bins in um
        )

    """
    # Filter out localizations not assigned to a trajectory, or 
    # trajectories that are length 1
    trajs = traj_length(trajs)
    T = trajs.loc[
        (trajs['trajectory']>=0) & (trajs['traj_len']>1), 
        :
    ].copy()

    # Add a column that indicates whether each localization is
    # the first in the corresponding trajectory (computed later)
    T['first_in_traj'] = 0

    # Order the trajectories by ascending frame / trajectory indices.
    # Trajectories are grouped together, and frames are ascending within
    # each trajectory group
    T = T.sort_values(by=['trajectory', 'frame'])

    # Take only the columns we care about in an ndarray. The operations
    # we will perform with ndarrays are much faster than their pandas
    # dataframe equivalents
    X = np.asarray(T[['frame', 'trajectory', 'y', 'x', 'first_in_traj']])

    # Determine which localizations are the first in their respective
    # trajectories
    X[0,4] = 1.0 
    X[1:,4] = ((X[1:,1] - X[:-1,1])>0).astype('float64')

    # Set up the binning scheme
    bin_edges = np.arange(0.0, max_disp+bin_size, bin_size)

    # Set up the output histogram
    H = np.zeros((n_intervals, bin_edges.shape[0]-1), dtype=np.int64)

    # In order to capture every possible displacement in the presence
    # of gaps, we need to check enough delays for the maximum interval
    # (n_intervals) with the maximum number of gaps between every
    # frame.
    n_delays = n_intervals * (n_gaps+1)

    for delay in range(1, n_delays+1):

        # Take the displacements
        D = X[delay:,:] - X[:-delay,:]

        # We're only interested in displacements between localizations
        # in the same trajectory
        in_same_traj = D[:,1] == 0

        # Check for displacements matching any of the desired frame
        # intervals
        for interval in range(1, n_intervals+1):

            if first_only:
                match = (D[:,0]==interval) & (D[:,4]==-1) & in_same_traj 
            else:
                match = (D[:,0]==interval) & in_same_traj 

            # Check if there are any displacements that match these
            # criteria
            if match.sum() > 0:

                # Get the radial displacements
                disps = np.sqrt((D[match,2:4]**2).sum(axis=1)) * pixel_size_um 

                # Bin these into a histogram
                H_int, _edges = np.histogram(disps, bins=bin_edges)

                # Accumulate into the histogram
                H[interval-1,:] = H[interval-1,:] + H_int

    # Return the histogram and the accompanying set of bin 
    # definitions
    return H, bin_edges

def get_max_gap(trajs):
    """
    Return the maximum gap present in a set of trajectories.

    args
    ----
        trajs       :   pandas.DataFrame

    returns
    -------
        int, the maximum gap present. For instance, 1 means
            that trajectories have no gaps (1 frame interval).

    """
    trajs = trajs.sort_values(by=['trajectory', 'frame'])
    X = np.asarray(trajs[['frame', 'trajectory']])
    D = X[1:,:] - X[:-1,:]
    in_same_traj = D[:,1]==0
    return int(D[in_same_traj, 0].max())

def get_spots_per_frame(trajs):
    """
    Return the number of spots per frame.

    Parameters
    ----------
        trajs       :   pandas.DataFrame, localizations

    Returns
    -------
        pandas.DataFrame, indexed by frame between 0 
            and the max frame index in *trajs*. The 
            column encoding the number of spots per 
            frame is "n_spots_per_frame".

    """
    if trajs.empty:
        return pd.DataFrame([], columns=["n_spots_per_frame"])
    result = pd.DataFrame(
        index=pd.Series(np.arange(0, trajs['frame'].max()+1)),
        columns=["n_spots_per_frame"]
    )

    # Calculate the number of spots per frame
    result["n_spots_per_frame"] = trajs.groupby("frame").size()

    # Account for frames that do not exist in the input dataframe
    result["n_spots_per_frame"] = result["n_spots_per_frame"].fillna(0)

    # Format as 64-bit integer
    result["n_spots_per_frame"] = result["n_spots_per_frame"].astype(np.int64)
    return result 

def filter_on_spots_per_frame(trajs, max_spots_per_frame=10,
    filter_kernel=21):
    """
    Mask a set of localizations by the total number of localizations
    in the corresponding frame.

    This function does the following:
        1. Take a set of localizations.
        2. Calculate the number of localizations in each frame.
        3. Smooth this signal with a uniform kernel of size 
            *filter_kernel*.
        4. Get the set of frames with total # of spots below
            *max_spots_per_frame*.
        5. Mark each localization in these frames with *True*,
            and otherwise with *False*.

    args
    ----
        trajs               :   pandas.DataFrame, localizations
        max_spots_per_frame :   int, the maximum number of spots
                                tolerated per frame
        filter_kernel       :   int, the size of the smoothing
                                kernel. If *None*, no smoothing is
                                performed.

    returns
    -------
        pandas.Series with index *trajs.index*. True if the 
            corresponding localization passed the filter.

    """
    # Calculate the number of spots per frame
    spots_per_frame = get_spots_per_frame(trajs)

    # Smooth the signal, if desired
    if not filter_kernel is None:
        spots_per_frame = uniform_filter(
            spots_per_frame.astype(np.float64),
            filter_kernel
        )

    # Get the set of frames that pass the filter
    acceptable = (spots_per_frame <= max_spots_per_frame).nonzero()[0]
    return trajs["frame"].isin(acceptable)















def get_values_as_in_dict(dict_to_copy: dict, summary_string: str) -> dict:
    print(summary_string)
    dict_to_return = {}
    for key, value in dict_to_copy.items():
        orig_cls = type(value)
        try:
            if orig_cls == bool:
                print("Enter %s (default is %s) : " % (key, value))
                input_v = input(
                    "Press y for true, n for false, anything else will be considered as default : "
                )
                if input_v == "y":
                    input_v = True
                elif input_v == "n":
                    input_v = False
                else:
                    print("Using default %s" % value)
                    input_v = value
            else:
                input_v = input("Enter %s (default is %s) : " % (key, value))
                if len(input_v.strip()) == 0:
                    print("Using default")
                    input_v = value
                else:
                    input_v = orig_cls(input_v)
        except:
            print(
                "Value entered %s is could not be converted to type %s"
                % (value, orig_cls)
            )
            print("Setting %s to default %s" % (key, value))
            input_v = value
        dict_to_return[key] = input_v
    return dict_to_return



from __future__ import annotations
import json
from typing import TYPE_CHECKING, Any, Callable, List
import os
import warnings
import pandas as pd
import numpy as np
import trackpy as tp
import logging
import napari
from glob import glob
import dask_image.imread
import dask.array as da
import dask_image.ndfilters
from skimage.filters import sato, threshold_local


from ..tif_tools.intensity import mean_intensity_center
from ..tif_tools.correct_drift import correct_drift
from ..tif_tools.localization import localize_movie

if TYPE_CHECKING:
    from .experiment import Experiment
    from ..processing.tif_pipeline import TifPipeline


class Acquisition:
    """
    An acquisition corresponds to a PALM movie.
    It is part of an :py:class:`Experiment`, and bound to a :py:class:`TifPipeline` with which it is processed.
    """

    def __init__(
        self,
        tif_file,
        experiment: Experiment,
        tif_pipeline: TifPipeline,
    ):
        self.tif_file = tif_file
        self.tif_pipeline = tif_pipeline
        self.experiment = experiment
        self.export_root = os.path.join(
            self.experiment.export_folder,
            self.tif_pipeline.name,
            ".".join(self.tif_file.split(".")[:-1]),
        )
        logging.debug(
            "export_root : file %s -> %s" % (self.tif_file, self.export_root)
        )
        export_parent_folder = os.sep.join(self.export_root.split(os.sep)[:-1])
        os.makedirs(export_parent_folder, exist_ok=True)

    @property
    def image(self) -> da.Array:
        """
        The actual movie, loaded with Dask.

        Returns:
            da.Array: the movie.
        """
        if not hasattr(self, "_image"):
            logging.info("Loading tif file %s" % self.tif_file)
            # self._image = tifffile.imread(
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._image = dask_image.imread.imread(
                    os.path.join(self.experiment.data_folder, self.tif_file),
                    nframes=300,
                )
        return self._image

    def get_property(self, col: str) -> Any:
        """Access an acquisition's property, read from its experiment's index table.

        Args:
            col (str): name of the index table column to look into

        Returns:
            Any: value of the corresponding row x column in the experiment's index table
        """
        df = self.experiment.index_df
        return df.loc[df.file == self.tif_file, col].values[0]

    @property
    def ID(self) -> str:
        return self.get_property("ID")

    def _clean_locs_table(self, df: pd.DataFrame) -> pd.DataFrame:
        for c in ["frame", "x", "y", "t"]:
            assert c in df.columns, "%s not in columns" % c
        all_cols = list(df.columns)
        for c in all_cols:
            if "Unnamed" in c:
                del df[c]
        assert (
            df.shape[0] == df.index.nunique()
        ), "There are duplicate indices in the locs table"
        return df

    @property
    def locs(self) -> pd.DataFrame:
        if not hasattr(self, "_locs"):
            try:
                df = pd.read_csv(self.locs_path)
                self._locs = self._clean_locs_table(df)
            except FileNotFoundError:
                print(
                    "This acquisition wasn't localized. Or perhaps with another Pipeline ?"
                )
                return None
        return self._locs

    @locs.setter
    def locs(self, value):
        if hasattr(self, "_raw_locs"):
            assert (
                self._raw_locs.shape[0] == value.shape[0]
            ), "Locs and raw_locs must have the same shape"
        value = self._clean_locs_table(value)
        value.to_csv(self.locs_path)
        self._locs = value

    @property
    def locs_path(self) -> str:
        if not hasattr(self, "_locs_path"):
            self._locs_path = self.export_root + ".locs"
        return self._locs_path

    @property
    def raw_locs(self) -> pd.DataFrame:
        if not hasattr(self, "_raw_locs"):
            raw_locs_path = self.raw_locs_path
            if os.path.exists(raw_locs_path):
                df = pd.read_csv(raw_locs_path)
                self._raw_locs = self._clean_locs_table(df)
            else:
                self.localize()
        return self._raw_locs

    @raw_locs.setter
    def raw_locs(self, value):
        value = self._clean_locs_table(value)
        value.to_csv(self.raw_locs_path)
        self._raw_locs = value

    @property
    def raw_locs_path(self) -> str:
        if not hasattr(self, "_raw_locs_path"):
            self._raw_locs_path = self.export_root + ".raw_locs"
        return self._raw_locs_path

    @property
    def intensity(self) -> pd.DataFrame:
        if not hasattr(self, "_intensity"):
            loaded = False
            intensity_path = self.intensity_path
            if os.path.exists(intensity_path):
                self._intensity = pd.read_csv(intensity_path)
                loaded = True
            if not loaded:
                self.compute_intensity()
        return self._intensity

    @property
    def intensity_path(self) -> str:
        if not hasattr(self, "_intensity_path"):
            self._intensity_path = self.export_root + ".int"
        return self._intensity_path

    @property
    def tubeness(self) -> np.array:
        if not hasattr(self, "_tubeness"):
            loaded = False
            tubeness_path = self.tubeness_path
            if os.path.exists(tubeness_path):
                self._tubeness = np.load(tubeness_path)
                loaded = True
            if not loaded:
                self.compute_tubeness()
        return self._tubeness

    @property
    def tubeness_path(self) -> str:
        if not hasattr(self, "_tubeness_path"):
            self._tubeness_path = self.export_root + ".tubeness"
        return self._tubeness_path

    @property
    def is_processed(self) -> bool:
        return (
            self.is_localized and self.is_tracked
        )  # and self.intensity_is_computed

    @property
    def is_localized(self) -> bool:
        return os.path.exists(self.raw_locs_path)

    @property
    def intensity_is_computed(self) -> bool:
        return os.path.exists(self.intensity_path)

    @property
    def tubeness_is_computed(self) -> bool:
        return os.path.exists(self.tubeness_path)

    @property
    def drift_is_corrected(self) -> bool:
        return os.path.exists(self.locs_path)

    @property
    def is_tracked(self):
        if not self.drift_is_corrected:
            return False
        locs_path = self.locs_path
        df = pd.read_csv(locs_path, nrows=10)
        return "n" in df.columns

    @property
    def polygon_folder(self) -> str:
        if not hasattr(self, "_acq_poly_path"):
            base = self.experiment.export_folder
            acq_ID = self.get_property("ID")
            poly_path = os.path.join(base, "polygons")
            if not os.path.exists(poly_path):
                os.mkdir(poly_path)
            self._acq_poly_path = os.path.join(poly_path, acq_ID)
            if not os.path.exists(self._acq_poly_path):
                os.mkdir(self._acq_poly_path)
        return self._acq_poly_path

    @property
    def polygon_files(self) -> List:
        return glob(os.path.join(self.polygon_folder, "*.polygon"))

    def add_columns_to_loc(self, columns: pd.DataFrame, save: bool = False):

        for c in columns.columns:
            if c in self._locs.columns:
                del self._locs[c]
            self._locs = self._locs.merge(
                columns[[c]], left_index=True, right_index=True, how="left"
            )
        if save:
            self._save_locs()

    def add_traj_cols_to_locs(self, traj_columns: pd.DataFrame):
        """
        traj_columns is a dataframe whose index corresponds to the 'n' column (traj ID)
        We merge it with the locs dataframe
        """
        to_delete_cols = [
            c for c in traj_columns.columns if c in self._locs.columns
        ]
        for col in to_delete_cols:
            del self._locs[col]
        self._locs = self._locs.merge(
            traj_columns, left_on="n", right_index=True, how="left"
        )

    def compute_intensity(self):
        if not self.intensity_is_computed:
            self._intensity = mean_intensity_center(self.image)
            self._save_intensity()
        else:
            logging.info("%s : intensity is already computed" % self.tif_file)

    def _save_intensity(self):
        storage_path = self.intensity_path
        self._intensity.to_csv(storage_path, index=False)
        logging.info(
            "Stored Intensity of ROI %s at path %s"
            % (self.tif_file, storage_path)
        )

    def compute_tubeness(self):
        if not self.tubeness_is_computed:
            self._compute_tubeness()
            self._save_tubeness()
        else:
            logging.info("%s : tubeness was already processed" % self.tif_file)

    def _compute_tubeness(self):

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mean = self.image[::10].mean(axis=0)
            mean_intensity = mean.compute()
            tubeness = sato(
                np.log10(mean_intensity), sigmas=3.0, black_ridges=False
            )
            tubeness_threshold = threshold_local(tubeness, 31)
            tubeness_mask = tubeness > tubeness_threshold
            self._tubeness = tubeness_mask

    def _save_tubeness(self):
        storage_path = self.tubeness_path
        np.save(storage_path, self._tubeness)
        logging.info("Saved tubeness at %s" % storage_path)

    def localize(self):
        if not self.is_localized:
            logging.info("Running localization for file %s" % self.tif_file)
            self._localize(**self.tif_pipeline.params.loc_params)
            self._save_raw_locs()
        else:
            logging.info(
                "Localization step has already been done for file %s"
                % self.tif_file
            )

    def _localize(
        self, threshold, sliding_window_filter, subpixel_mode, **args
    ):

        data = self.image
        self._raw_locs = localize_movie(
            data,
            progress_bar=True,
            verbose=False,
            sliding_filter=sliding_window_filter,
            factor=threshold,
            subpixel_mode=subpixel_mode,
        )
        self._raw_locs["t"] = self._raw_locs["frame"] * self.experiment.DT
        self._raw_locs[
            ["x", "y"]
        ] *= self.experiment.pixel_size  # x, y are in um
        N = self._raw_locs.shape[0]
        T = data.shape[0]
        logging.info("==== %d locs\t %.2f / frame" % (N, N / T))

    def _save_raw_locs(self):
        storage_path = self.raw_locs_path
        self._raw_locs.to_csv(storage_path, index=False)
        logging.info(
            "Stored raw locs of ROI %s at path %s"
            % (self.tif_file, storage_path)
        )

    def correct_drift(self):
        if not self.drift_is_corrected:
            if self.tif_pipeline.params.loc_params["correct_drift"] == True:
                self._correct_drift()
            else:
                self._locs = self.raw_locs
            self._save_locs()
        else:
            logging.info(
                "Drift correction step has already been done for file %s"
                % self.tif_file
            )

    def _correct_drift(self):
        logging.info("Correcting drift iteration #1/2")
        df, self._shift_1 = correct_drift(self.raw_locs, L=0.2, step_size=0.03)
        logging.info("Correcting drift iteration #2/2")
        self._locs, self._shift_2 = correct_drift(df, L=0.05, step_size=0.01)

    def _save_locs(self):
        storage_path = self.locs_path
        self._locs.to_csv(storage_path, index=False)
        logging.info(
            "Stored locs of ROI %s at path %s" % (self.tif_file, storage_path)
        )

    def track(self):
        if not self.is_tracked:
            logging.info("Running tracking for file %s" % self.tif_file)
            self._track(**self.tif_pipeline.params.track_params)
            self._save_locs()
        else:
            logging.info(
                "Tracking was already done for file %s" % self.tif_file
            )

    def _track(self, mode, max_diffusivity):

        if mode == "trackpy":
            self._track_trackpy(max_diffusivity)
        else:
            raise "Unknown tracking mode %s" % mode

    def _track_trackpy(self, max_diffusivity):
        max_radius = np.sqrt(2 * max_diffusivity * self.experiment.DT)
        logging.info("Starting to track %d localisations" % self.locs.shape[0])
        logging.info("Max radius is %.2f" % max_radius)

        tracks = tp.link(
            self.locs, search_range=max_radius, link_strategy="drop"
        )
        self._locs["n"] = tracks["particle"]
        traj_length = self._locs.n.value_counts()
        n_long_trajs = (traj_length >= 7).sum()
        logging.info(
            "Found %d long trajectories for file %s"
            % (n_long_trajs, self.tif_file)
        )

    # EXPLOITATION OF TRACKS AND LOCS

    def trajectories_list(
        self,
        min_length: int = 7,
        return_indices: bool = True,
        filter: Callable = None,
    ):
        """
        Returns a list of trajectories whose length is above a given threshold, possibly filtered according to the locs they're based on

        Args:
            min_length (int, optional): Defaults to 7.
            return_indices (bool, optional): Whether to return indices ('n' column of the locs DataFrame) along with coordinates. Defaults to True.
            filter (Callable, optional): Callable, which takes as input the locs DataFrame and returns a boolean Series with the same index. Defaults to None.

        Returns:
            _type_: either a list of trajectories, or a tuple containing this same list and the list of indices
        """
        if filter is not None:
            df = pd.DataFrame(index=self.locs.index)
            df["n"] = self.locs.n
            df["cond"] = filter(self.locs)
            # The filter is applied point by point
            # Then we make sure to only select trajectories fo which all points pass the filter
            cond = df.groupby("n")["cond"].min()
            cond = self.locs.n.isin(cond[cond].index.tolist())
        else:
            cond = ~self.locs.x.isnull()
        traj_length = self.locs.loc[cond].n.value_counts()
        good_trajs = traj_length.loc[traj_length >= min_length].index.tolist()
        indices = []
        trajs = []
        for n, traj in (
            self.locs.loc[self.locs.n.isin(good_trajs)]
            .sort_values(["n", "t"])
            .groupby("n")
        ):
            indices.append(n)
            trajs.append(traj[["x", "y"]].values)
        to_return = (trajs,)
        if return_indices:
            to_return = to_return + (indices,)
        return to_return

    def get_traj(self, n: int):
        return (
            self.locs.loc[self.locs.n == n].sort_values("t")[["x", "y"]].values
        )

    def basic_stats(self):
        stats = {}
        stats["file"] = self.tif_file
        stats["n_frames"] = self.locs.frame.max()
        stats["duration"] = stats["n_frames"] * self.experiment.DT
        # the above is not exactly true, but is much faster than what's below
        # self.image.shape[0]
        if not self.is_processed:
            return stats
        stats["n_locs"] = self.locs.shape[0]
        traj_length = self.locs.n.value_counts()
        for L in [4, 7, 10, 15, 25]:
            stats["n_trajs_%d" % L] = (traj_length >= L).sum()
        return stats

    # VIEWER FUNCTIONS

    def view(
        self,
        min_traj_length=1,
        polygon_ID_col: str = None,
        short_for_tests: bool = False,
        contrast_limits: tuple = (100, 500),
    ):
        viewer = napari.viewer.Viewer(title=os.path.dirname(self.tif_file))
        viewer.add_image(
            self.image[:1000] if short_for_tests else self.image,
            name=self.tif_file.split("/")[-1],
            multiscale=False,
            contrast_limits=contrast_limits,
            scale=(1, self.experiment.pixel_size, self.experiment.pixel_size),
        )
        if self.locs is not None:
            self.view_points(viewer, polygon_ID_col=polygon_ID_col)
            self.view_tracks(
                viewer,
                min_length=min_traj_length,
                polygon_ID_col=polygon_ID_col,
            )
            self.view_polygons(viewer)

    def view_points(
        self, viewer=None, polygon_ID_col: str = None, subsample: int = None
    ):
        if viewer is None:
            viewer = napari.viewer.Viewer()
        locs = self.locs.copy()
        if subsample is not None and subsample < locs.shape[0]:
            locs = locs.sample(subsample)

        points = locs[["frame", "x", "y"]].copy()
        # points[["x","y"]] /= self.pixel_size

        props = {"frame": points["frame"].values}
        for c in ["abs_density", "norm_density"]:
            if c in locs.columns:
                props[c] = locs[c].values
                logging.info("Adding property %s" % c)
        if polygon_ID_col is not None and polygon_ID_col in locs.columns:
            props[polygon_ID_col] = pd.Categorical(
                locs[polygon_ID_col].fillna("no")
            ).codes
            logging.info("Adding polygon tags")

        viewer.add_points(
            points,
            size=0.15,
            face_color="frame",
            symbol="x",
            properties=props,
            edge_width=0.03,
            # edge_color="#dd3333",
            name="locs (adjusted)",
        )

        points = self.raw_locs.loc[locs.index, ["frame", "x", "y"]].copy()
        # points[["x","y"]] /= self.pixel_size
        viewer.add_points(
            points,
            size=0.15,
            symbol="x",
            face_color="frame",
            edge_width=0.03,
            # edge_color="#3333dd",
            name="locs (unadjusted)",
            properties=props,
            visible=False,
        )

    # FOR THE VIEWER

    def view_tracks(
        self, viewer=None, min_length=1, polygon_ID_col: str = None
    ):
        if viewer is None:
            viewer = napari.viewer.Viewer()

        traj_length = self.locs.n.value_counts()
        good_trajs = traj_length.loc[traj_length >= min_length].index.tolist()

        tracks = self.locs.loc[
            self.locs.n.isin(good_trajs),
        ].copy()
        # tracks = self.compute_tracks_displacement(tracks)
        # tracks[["x","y"]] /= self.pixel_size
        tracks["t"] = tracks["frame"].astype(int)
        props = {"time": tracks["t"].values}
        if polygon_ID_col is not None and polygon_ID_col in tracks.columns:
            props[polygon_ID_col] = pd.Categorical(
                tracks[polygon_ID_col].fillna("no")
            ).codes
        viewer.add_points(
            tracks[["t", "x", "y"]],
            size=0.1,
            symbol="x",
            # face_color="#ffffff00",
            edge_width=0.03,
            edge_color="time",
            name="... used in tracks",
            properties=props,
        )
        track_properties = {
            "length": np.clip(
                pd.merge(
                    tracks,
                    tracks.groupby("n")[["x"]]
                    .count()
                    .rename(columns={"x": "length"}),
                    left_on="n",
                    right_index=True,
                )["length"].values,
                a_min=7,
                a_max=30,
            ),
            # "log_D": np.log10(
            #    pd.merge(
            #        tracks[["n"]],
            #        tracks.loc[~tracks.is_last].groupby("n")[["dr_2"]].mean(),
            #        left_on="n",
            #        right_index=True,
            #    )["dr_2"].values
            # ),
            # "log_dr": np.log10(tracks["dr_2"].values),
        }
        track_properties.update(props)

        viewer.add_tracks(
            tracks[["n", "t", "x", "y"]],
            tail_width=0.2,
            name="Tracks",
            tail_length=10,
            head_length=10,
            properties=track_properties,
        )

    def view_polygons(self, viewer: napari.Viewer = None):
        if viewer is None:
            viewer = napari.viewer.Viewer()
        polygons = []

        for p in self.polygon_files:
            polygon_dict = json.load(open(p, "r"))
            if len(polygon_dict["x"]) < 3:
                logging.debug("%s a moins de 3 points, on le supprime" % p)
                os.remove(p)
                continue
            synapse_ID = os.path.split(p)[1].split(".")[0]
            polygon_path = np.stack(
                [(x, y) for x, y in zip(polygon_dict["x"], polygon_dict["y"])]
            )
            polygons.append(polygon_path)

        if len(polygons) > 0:
            viewer.add_shapes(
                data=polygons,
                shape_type="polygon",
                face_color_cycle=["#AA77CC44", "#77AACC44", "#CC77AA44"],
                edge_width=0.1,
                edge_color="red",
            )



import numpy as np
from scipy.ndimage.morphology import binary_dilation
from skimage.filters import gaussian, threshold_otsu
from skimage.morphology import closing, square, remove_small_objects, convex_hull_object
from skimage.segmentation import clear_border
from skimage.measure import label
import pandas as pd
from scipy.stats import binned_statistic_2d
import matplotlib.pyplot as plt


def get_cell_mask(data: np.ndarray):
    data_max = .5*(np.max(data[::10], axis=0) + np.mean(data[::10], axis=0))
    gblur = gaussian(data_max, sigma=4)
    threshold = threshold_otsu(gblur)
    #plt.figure()
    #plt.imshow(gblur)
    bw = closing(gblur > threshold, square(4))
    #plt.figure()
    #plt.imshow(bw)
    extra_width = 5
    bw = binary_dilation(bw, iterations=extra_width)
    #plt.figure()
    #plt.imshow(bw)

    cleared = remove_small_objects(bw, 100)
    #plt.figure()
    #plt.imshow(cleared)
    convex_image = convex_hull_object(cleared, connectivity=2)
    label_image = label(convex_image).astype(int)
    
    return label_image


def tag_localizations_per_cell(pos: pd.DataFrame, labels: np.ndarray, scale: float):
    XY = pos[["x", "y"]].values/scale
    cell = pd.Series(data=np.zeros(pos.shape[0]))
    _, _, _, bins = binned_statistic_2d(x=XY[:, 0], y=XY[:, 1], values=np.ones_like(XY[:,0]), bins=(
        np.arange(labels.shape[0]+1), np.arange(labels.shape[1]+1)))
    flat_labels = np.reshape(labels, (-1,))
    for l in np.unique(labels):
        if l == 0:
            continue
        good_bins = np.where(flat_labels == l)
        cell.loc[np.isin(bins,good_bins)] = l
    return cell
        


#!/usr/bin/env python
"""
subpixel.py -- localize PSFs to subpixel resolution

"""
# Numeric
import numpy as np

# Dataframes
import pandas as pd

# Image processing
from scipy import ndimage as ndi

# Low-level subpixel localization utilities
from .helper import (
    assign_methods,
    ring_mean,
    rs,
    I0_is_crazy,
    estimate_I0,
    estimate_I0_multiple_points,
    estimate_snr,
    fit_ls_int_gaussian,
    fit_ls_point_gaussian,
    fit_poisson_int_gaussian,
    check_2d_gauss_fit,
)


def centroid(I, sub_bg=False):
    """
    Find the center of a spot by taking its center
    of mass.

    args
    ----
        I       :   2D ndarray (YX), spot image
        sub_bg  :   bool, subtract background prior to
                    taking center of mass

    returns
    -------
        dict {
            y           : y centroid (pixels),
            x           : x centroid (pixels),
            bg          : estimated background intensity per
                            pixel (AU),
            I0          : integrated spot intensity above
                            background (AU),
            error_flag  : int, the error code. 0 indicates
                            no errors.
        }

    """
    # Estimate background
    bg = ring_mean(I)

    # Subtract background
    I_sub = np.clip(I - bg, 0.0, np.inf)

    # Integrated intensity above background
    I0 = I_sub.sum()

    # Find spot centers
    if sub_bg:
        y, x = ndi.center_of_mass(I_sub)
    else:
        y, x = ndi.center_of_mass(I)

    # Estimate SNR
    snr = estimate_snr(I, I0)

    # Return parameter estimates
    return dict(
        (
            ("y", y),
            ("x", x),
            ("I0", I0),
            ("bg", bg),
            ("error_flag", 0),
            ("snr", snr),
        )
    )


def radial_symmetry(I, sigma=1.0, **kwargs):
    """
    Estimate the center of a spot by the radial symmetry
    method, described in Parthasarathy et al. Nat Met 2012.

    Also infer the intensity of the spots assuming an
    integrated Gaussian PSF. This is useful as a first guess
    for iterative localization techniques.

    args
    ----
        I       :   2D ndarray, spot image
        sigma   :   float, static integrated 2D Gaussian
                    width

    returns
    -------
        dict {
            y   :    estimated y center (pixels),
            x   :    estimated x center (pixels),
            I0  :    estimated PSF intensity (AU),
            bg  :    estimated background intensity per
                        pixel (AU),
            error_flag  :   int, the error flag. 0
                            if there are no errors,
        }

    """
    # Estimate spot centers
    y, x = rs(I)

    # Estimate background level
    bg = ring_mean(I)

    # Estimate Gaussian intensity
    I0 = estimate_I0_multiple_points(I, y, x, bg, sigma=sigma)

    # If I0 is crazy, use a different guess
    if I0_is_crazy(I0):
        I0 = np.clip(I - bg, 0.0, np.inf).sum()
        error_flag = 1
    else:
        error_flag = 0

    # Estimate SNR
    snr = estimate_snr(I, I0)

    # Return parameter estimate
    return dict(
        (
            ("y", y),
            ("x", x),
            ("I0", I0),
            ("bg", bg),
            ("error_flag", error_flag),
            ("snr", snr),
        )
    )


def ls_int_gaussian(
    I,
    sigma=1.0,
    ridge=0.0001,
    max_iter=10,
    damp=0.3,
    convergence=1.0e-4,
    divergence=1.0,
):
    """
    Estimate the maximum likelihood parameters for a symmetric
    2D integrated Gaussian PSF model, given an observed spot
    *I* with normally-distributed noise.

    This method uses radial symmetry for a first guess, followed
    by a Levenberg-Marquardt routine for refinement. The core
    calculation is performed in quot.helper.fit_ls_int_gaussian.

    args
    ----
        I           :   2D ndarray (YX), the observed spot
        sigma       :   float, static Gaussian width
        ridge       :   float, initial regularization term
                        for inversion of the Hessian
        max_iter    :   int, the maximum tolerated number
                        of iterations
        damp        :   float, damping factor for the update
                        at each iteration. Larger means faster
                        convergence, but less stable.
        convergence :   float, maximum magnitude of the update
                        vector at which to call convergence.
        divergence  :   float, divergence criterion

    returns
    -------
        dict, the parameter estimate, error estimates, and
            related parameters about the fitting problem
            (see below). Some of these can be useful for
            QC.

    """
    # Make the initial guess by radial symmetry
    guess = np.array([*rs(I), 0.0, ring_mean(I)])
    guess[2] = estimate_I0(I, guess[0], guess[1], guess[3], sigma=sigma)

    # Run the fitting routine
    pars, err, H_det, rmse, n_iter = fit_ls_int_gaussian(
        I,
        guess,
        sigma=sigma,
        ridge=ridge,
        max_iter=max_iter,
        damp=damp,
        convergence=convergence,
        divergence=divergence,
    )

    # Check for crazy fits
    if check_2d_gauss_fit(I.shape, pars):
        error_flag = 0
    else:
        error_flag = 1

    # Estimate SNR
    snr = estimate_snr(I, pars[2])

    # Return parameter estimate
    return dict(
        (
            ("y", pars[0]),
            ("x", pars[1]),
            ("I0", pars[2]),
            ("bg", pars[3]),
            ("y_err", err[0]),
            ("x_err", err[1]),
            ("I0_err", err[2]),
            ("bg_err", err[3]),
            ("H_det", H_det),
            ("error_flag", error_flag),
            ("snr", snr),
            ("rmse", rmse),
            ("n_iter", n_iter),
        )
    )


def ls_point_gaussian(
    I,
    sigma=1.0,
    ridge=0.0001,
    max_iter=10,
    damp=0.3,
    convergence=1.0e-4,
    divergence=1.0,
):
    """
    Estimate the maximum likelihood parameters for a symmetric
    2D pointwise-evaluated Gaussian PSF model, given an observed
    spot *I* with normally-distributed noise.

    Both integrated and pointwise-evaluated models have the same
    underlying model: a symmetric, 2D Gaussian PSF. The
    distinction is how they handle sampling on discrete pixels:

        - the point Gaussian takes the value on each pixel to
            be equal to the PSF function evaluated at the center
            of that pixel

        - the integrated Gaussian takes the value on each pixel
            to be equal to the PSF integrated across the whole
            area of the pixel

    Integrated Gaussian models are more accurate and less prone
    to edge biases, at the cost of increased complexity and
    perhaps lower speed.

    This method uses radial symmetry for a first guess, followed
    by a Levenberg-Marquardt routine for refinement. The core
    calculation is performed in quot.helper.fit_ls_point_gaussian.

    args
    ----
        I           :   2D ndarray (YX), the observed spot
        sigma       :   float, static Gaussian width
        ridge       :   float, initial regularization term
                        for inversion of the Hessian
        max_iter    :   int, the maximum tolerated number
                        of iterations
        damp        :   float, damping factor for the update
                        at each iteration. Larger means faster
                        convergence, but less stable.
        convergence :   float, maximum magnitude of the update
                        vector at which to call convergence.
        divergence  :   float, divergence criterion

    returns
    -------
        dict, the parameter estimate, error estimates, and
            related parameters about the fitting problem
            (see below). Some of these can be useful for
            QC.

    """
    # Make the initial guess by radial symmetry
    guess = np.array([*rs(I), 0.0, ring_mean(I)])
    guess[2] = estimate_I0(I, guess[0], guess[1], guess[3], sigma=sigma)

    # Run the fitting routine
    pars, err, H_det, rmse, n_iter = fit_ls_point_gaussian(
        I,
        guess,
        sigma=sigma,
        ridge=ridge,
        max_iter=max_iter,
        damp=damp,
        convergence=convergence,
        divergence=divergence,
    )

    # Check for crazy fits
    if check_2d_gauss_fit(I.shape, pars):
        error_flag = 0
    else:
        error_flag = 1

    # Estimate SNR
    snr = estimate_snr(I, pars[2])

    # Return parameter estimate
    return dict(
        (
            ("y", pars[0]),
            ("x", pars[1]),
            ("I0", pars[2]),
            ("bg", pars[3]),
            ("y_err", err[0]),
            ("x_err", err[1]),
            ("I0_err", err[2]),
            ("bg_err", err[3]),
            ("H_det", H_det),
            ("error_flag", error_flag),
            ("snr", snr),
            ("rmse", rmse),
            ("n_iter", n_iter),
        )
    )


def poisson_int_gaussian(
    I,
    sigma=1.0,
    ridge=0.0001,
    max_iter=10,
    damp=0.3,
    convergence=1.0e-4,
    divergence=1.0,
):
    """
    Estimate the maximum likelihood parameters for a symmetric
    2D integrated Gaussian PSF model, given an observed spot
    *I* with Poisson-distributed noise.

    This method uses radial symmetry for a first guess, followed
    by a Levenberg-Marquardt routine for refinement. The core
    calculation is performed in quot.helper.fit_poisson_int_gaussian.

    args
    ----
        I           :   2D ndarray (YX), the observed spot
        sigma       :   float, static Gaussian width
        ridge       :   float, initial regularization term
                        for inversion of the Hessian
        max_iter    :   int, the maximum tolerated number
                        of iterations
        damp        :   float, damping factor for the update
                        at each iteration. Larger means faster
                        convergence, but less stable.
        convergence :   float, maximum magnitude of the update
                        vector at which to call convergence.
        divergence  :   float, divergence criterion

    returns
    -------
        dict, the parameter estimate, error estimates, and
            related parameters about the fitting problem
            (see below). Some of these can be useful for
            QC.

    """
    # Make the initial guess by radial symmetry
    guess = np.array([*rs(I), 0.0, ring_mean(I)])
    guess[2] = estimate_I0(I, guess[0], guess[1], guess[3], sigma=sigma)

    # Run the fitting routine
    pars, err, H_det, rmse, n_iter = fit_poisson_int_gaussian(
        I,
        guess,
        sigma=sigma,
        ridge=ridge,
        max_iter=max_iter,
        damp=damp,
        convergence=convergence,
        divergence=divergence,
    )

    # Check for crazy fits
    if check_2d_gauss_fit(I.shape, pars):
        error_flag = 0
    else:
        error_flag = 1

    # Estimate SNR
    snr = estimate_snr(I, pars[2])

    # Return parameter estimate
    return dict(
        (
            ("y", pars[0]),
            ("x", pars[1]),
            ("I0", pars[2]),
            ("bg", pars[3]),
            ("y_err", err[0]),
            ("x_err", err[1]),
            ("I0_err", err[2]),
            ("bg_err", err[3]),
            ("H_det", H_det),
            ("error_flag", error_flag),
            ("snr", snr),
            ("rmse", rmse),
            ("n_iter", n_iter),
        )
    )


##########################################
## MAIN SUBPIXEL LOCALIZATION FUNCTIONS ##
##########################################

# All localization methods available
METHODS = {
    "centroid": centroid,
    "radial_symmetry": radial_symmetry,
    "ls_int_gaussian": ls_int_gaussian,
    "ls_point_gaussian": ls_point_gaussian,
    "poisson_int_gaussian": poisson_int_gaussian,
}

# Wrapper for all localization methods to run on
# a single PSF.

# Example usage:
#  fit_pars = localize(psf_img, method='ls_int_gaussian',
#     **method_kwargs)
localize = assign_methods(METHODS)(lambda I, **kwargs: None)

# Wrapper for all localization methods on a frame
# with several detections.
def localize_frame(
    img,
    positions,
    method=None,
    window_size=9,
    camera_bg=0.0,
    camera_gain=1.0,
    **method_kwargs
):
    """
    Run localization on multiple spots in a large 2D image,
    returning the result as a pandas DataFrame.

    args
    ----
        img         :   2D ndarray (YX), the image frame
        positions   :   2D ndarray of shape (n_spots, 2),
                        the y and x positions at which to
                        localize spots
        method      :   str, a method in METHODS
        window_size :   int, the fitting window size
        camera_bg   :   float, the BG per subpixel in the camera
        camera_gain :   float, the camera gain (grayvalues/photon)
        method_kwargs:  to the localization method

    returns
    -------
        pandas.DataFrame with columns ['y_detect', 'x_detect']
            plus all of the outputs from the localization function

    """
    # If no method passed, return the unmodified dataframe
    if method is None:
        return positions

    # Remove detections too close to the edge for a square
    # subwindow
    if len(positions.shape) == 2:
        hw = window_size // 2
        positions = positions[
            (positions[:, 0] >= hw)
            & (positions[:, 0] < img.shape[0] - hw)
            & (positions[:, 1] >= hw)
            & (positions[:, 1] < img.shape[1] - hw),
            :,
        ]

        # Get the localization method
        method_f = METHODS.get(method)

        # Localize a PSF in a subwindow of the image
        def localize_subwindow(yd, xd):
            psf_img = (
                np.clip(
                    img[yd - hw : yd + hw + 1, xd - hw : xd + hw + 1]
                    - camera_bg,
                    0,
                    np.inf,
                )
                / camera_gain
            )
            r = method_f(psf_img, **method_kwargs)
            r.update({"y_detect": yd, "x_detect": xd})
            r["y"] = r["y"] + yd - hw
            r["x"] = r["x"] + xd - hw
            return r

        # Run localization on all PSF subwindows
        result = pd.DataFrame(
            [localize_subwindow(yd, xd) for yd, xd in positions]
        )
        return result

    else:
        return pd.DataFrame([])



from __future__ import annotations
from abc import ABC, abstractmethod, abstractproperty
from ast import arg
import logging
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import dask.array as da
import dask.dataframe as dd
from dask import delayed
from dask.diagnostics import ProgressBar

from ...data_structure.acquisition import Acquisition
from ...data_structure.experiment import Experiment


class ProcessingStep(ABC):

    widget_options = {}
    widget_types = {}

    @abstractproperty
    def name(self):
        return "Abstract step"

    def __str__(self):
        desc = "%s (%s) :\n" % (self.name, self.__class__.__qualname__)
        for param, value in self.__dict__.items():
            desc += "\t\t %s : %s\n" % (param, value)
        return desc

    @abstractmethod
    def process(self, *args):
        """
        Depending on their types, subclasses might have different effector functions.
        e.g. Trackers'effector function is named ``.track()``
        This function is just meant to be a link towards the effector function
        """
        pass

    def to_dict(self) -> dict:
        params_dict = dict(self.__dict__)
        all_keys = list([key for key in params_dict])
        for key in all_keys:
            if key[0] == "_":
                # Do not include parameters starting with _
                del params_dict[key]

        return {self.__class__.__qualname__: params_dict}

    def update_param(self, param, *args):
        logging.debug("Updating %s" % param)
        value = args[0]
        logging.debug("%s -> %s" % (self.__dict__[param], value))
        self.__dict__[param] = value

    @property
    def is_localizer(self):
        # To be overriden in localizers and detectors
        return False

    @property
    def is_detector(self):
        # To be overriden in detectors
        return False


class Tracker(ProcessingStep):
    @abstractmethod
    def track(self, locs: pd.DataFrame) -> pd.DataFrame:
        locs["n"] = np.arange(locs.shape[0])
        return locs

    @property
    def name(self):
        return "Tracker"

    @property
    def action_name(self):
        return "Track"

    def process(self, *args):
        return self.track(*args)

    def estimate_delta_t(self, locs):
        delta_t = (locs["t"].max() - locs["t"].min()) / (
            locs["frame"].max() - locs["frame"].min()
        )
        logging.info("Tracker %s estimated delta_t = %.3f" % (self, delta_t))
        return delta_t


class MoviePreProcessor(ProcessingStep):
    @abstractmethod
    def preprocess(self, mov: da.Array) -> pd.DataFrame:
        return mov

    @property
    def name(self):
        return "Movie Processor"

    def process(self, *args):
        logging.debug("%s is about to process" % self.name)
        return self.preprocess(*args)


class Detector(ProcessingStep):

    cols_dtype = {
        "x": float,
        "y": float,
        "frame": int,
    }

    @abstractmethod
    def detect_frame(self, img: np.array) -> pd.DataFrame:
        """Detect spots from a temporal slice of an image. Override in subclasses

        Args:
            img (np.array): 3D array [T, X, Y]

        Returns:
            pd.DataFrame: localizations table with the following columns : x, y, frame
        """
        return pd.DataFrame.from_dict(
            {"x": [], "y": [], "frame": []}, orient="columns"
        )

    def movie_detection(self, mov: da.Array):
        slice_size = mov.chunksize[0]
        n_slices = mov.shape[0] // slice_size
        positions_dfs = []
        for i in range(n_slices + 1):
            start = i * slice_size
            end = min((i + 1) * slice_size, mov.shape[0])
            if start >= end:
                continue
            positions_dfs.append(
                delayed(self.detect)(
                    mov[start:end],
                    frame_start=start,
                )
            )
        loc_results_delayed = dd.from_delayed(
            positions_dfs,
            verify_meta=False,
            meta=self.cols_dtype,
        )
        with ProgressBar():
            # with warnings.catch_warnings():
            # warnings.simplefilter("ignore", category="RuntimeWarning")
            loc_results = loc_results_delayed.compute()
        loc_results.set_index(np.arange(loc_results.shape[0]), inplace=True)
        loc_results["detection_index"] = np.arange(loc_results.shape[0])
        return loc_results

    def detect(self, img: np.array, frame_start: int = 0) -> pd.DataFrame:
        detections = []
        for frame_idx in range(img.shape[0]):
            frame = img[frame_idx].T
            data = self.detect_frame(frame)
            if data.shape[0] > 0:
                frame_detections = pd.DataFrame(data=data, columns=["x", "y"])
                frame_detections["frame"] = frame_idx
                # Localize spots to subpixel resolution
                detections.append(frame_detections)

        if len(detections) > 0:
            detections = pd.concat(detections, ignore_index=True, sort=False)
        else:
            detections = pd.DataFrame.from_dict(
                {"x": [], "y": [], "frame": []}
            )
        # Performs checks on the returned pd.DataFrame
        detections["frame"] += frame_start
        for c, v in self.cols_dtype.items():
            assert c in detections.columns, "%s not in columns" % c
        for c in self.cols_dtype:
            if c in detections.columns:
                detections[c] = detections[c].astype(self.cols_dtype[c])
        return detections

    @property
    def name(self):
        return "Abstract Detector"

    @property
    def action_name(self):
        return "Detect spots"

    def process(self, *args):
        return self.movie_detection(*args)

    @property
    def is_localizer(self):
        return True

    @property
    def is_detector(self):
        return True


class SubpixelLocalizer(ProcessingStep):

    cols_dtype = {
        "x": float,
        "y": float,
        "frame": int,
    }

    @abstractmethod
    def localize_frame(
        self, img: np.array, detections: np.array
    ) -> pd.DataFrame:
        """
        Extract localizations from a slice of a .tif file. Override in subclasses

        Args:
            img (np.array): 2D array [X, Y]

            detections (np.array):  2D array [X, Y] center of spots (pixel indices)

        Returns:
            pd.DataFrame: localizations table with the following columns : x, y (in pixel units)
        """
        return pd.DataFrame.from_dict({"x": [], "y": []}, orient="columns")

    def movie_localization(self, mov: da.Array, detections: pd.DataFrame):
        slice_size = mov.chunksize[0]
        n_slices = mov.shape[0] // slice_size
        positions_dfs = []
        for i in range(n_slices + 1):
            start = i * slice_size
            end = min((i + 1) * slice_size, mov.shape[0])
            if start >= end:
                continue
            positions_dfs.append(
                delayed(self.localize)(
                    mov[start:end],
                    frame_start=start,
                    detections=detections.loc[
                        (detections.frame >= start) & (detections.frame < end)
                    ],
                )
            )
        loc_results_delayed = dd.from_delayed(
            positions_dfs,
            verify_meta=False,
            meta=self.cols_dtype,
        )
        with ProgressBar():
            # with warnings.catch_warnings():
            # warnings.simplefilter("ignore", category="RuntimeWarning")
            loc_results = loc_results_delayed.compute()
        loc_results.set_index(np.arange(loc_results.shape[0]), inplace=True)
        return loc_results

    def localize(
        self, img: np.array, detections: pd.DataFrame, frame_start: int = 0
    ) -> pd.DataFrame:
        locs = []
        for i in range(img.shape[0]):
            frame = img[i]
            frame_locs = self.localize_frame(
                frame,
                np.array(
                    detections.loc[
                        detections.frame == i + frame_start, ["x", "y"]
                    ].values,
                    dtype=int,
                ),
            )
            frame_locs["frame"] = i
            locs.append(frame_locs)
        # Performs checks on the returned pd.DataFrame
        locs = pd.concat(locs, ignore_index=True)
        if locs.shape[0] > 0:
            locs["frame"] += frame_start
            for c, v in self.cols_dtype.items():
                assert c in locs.columns, "%s not in columns" % c
            for c in self.cols_dtype:
                if c in locs.columns:
                    locs[c] = locs[c].astype(self.cols_dtype[c])
        return locs

    @property
    def name(self):
        return "Abstract Localizer"

    @property
    def action_name(self):
        return "Localize"

    def process(self, *args):
        return self.movie_localization(*args)

    @property
    def is_localizer(self):
        # To be overriden in localizers
        return True


class LocProcessor(ProcessingStep):
    @abstractmethod
    def process(
        self,
        mov: da.Array,
        locs: pd.DataFrame,
        pixel_size: float,
    ) -> pd.DataFrame:
        return locs

    @property
    def name(self):
        return "Locs processor doing nothing"

    def process(self, *args):
        return self.process(*args)



from .processing import TifPipelineWidget
from .processing import TifPipeline
from .processing.steps import *
from napari import Viewer


class PalmariWidget(TifPipelineWidget):
    def __init__(self, napari_viewer: Viewer):
        tp = TifPipeline.from_dict(
            {
                "name": "palmari-default",
                "detector": {"BaseDetector": {"k": 1.5, "w": 21, "t": 1.0}},
                "localizer": {
                    "MaxLikelihoodLocalizer": {
                        "method": "ls_int_gaussian",
                        "window_size": 9,
                        "sigma": 1.5,
                    }
                },
                "tracker": {
                    "DiffusionTracker": {
                        "max_diffusivity": 5.0,
                        "y_diff": 0.9,
                        "init_cost": 50.0,
                        "d_bound_naive": 0.1,
                    }
                },
            }
        )
        print(tp)
        super().__init__(tif_pipeline=tp, napari_viewer=napari_viewer)



#!/usr/bin/env python
"""
__init__.py

ALL THE CODE IN THIS FOLDER WAS COPIED/ADAPTED FROM 
https://github.com/alecheckert/quot
Which is under MIT LICENSE

"""
# Enforce PySide2 backend to pyqtgraph
# import PySide2
# import pyqtgraph

# Core functions to run localization and tracking on single files
# or directories
from .core import (
    localize_file,
    track_file,
    track_directory,
    retrack_file,
    retrack_files,
)

# Read and filter image files
from .read import ImageReader, read_config
from .chunkFilter import ChunkFilter

# Find spots
from .findSpots import detect

# Localize spots to subpixel resolution
from .subpixel import localize, localize_frame

# Reconnection spots into trajectories
from .track import track



import logging
from scipy.interpolate import interp1d
import numpy as np
from tqdm import tqdm
import pandas as pd
from dask import delayed
import dask.array as da
from skimage.registration import phase_cross_correlation


def get_optimal_shift(pos1, pos2, L, step):
    # objective : find dx, dy to move pos2 on pos1
    best = (None, None)
    bins_x = np.arange(pos1.min() - L, pos1.max() + L, step)
    bins_y = np.arange(pos2.min() - L, pos2.max() + L, step)

    hist1, _, _ = np.histogram2d(
        pos1[:, 0], pos1[:, 1], bins=(bins_x, bins_y), normed=True
    )
    hist2, _, _ = np.histogram2d(
        pos2[:, 0], pos2[:, 1], bins=(bins_x, bins_y), normed=True
    )

    # subpixel precision
    shift, error, diffphase = phase_cross_correlation(hist1, hist2)
    return shift * step
    """
    best_correlation = 0.0

    best = np.zeros(2)

    for dx in np.arange(-L, L, step):
        for dy in np.arange(-L, L, step):
            pos2_d = pos2 - np.array([dx, dy])

            hist2, _, _ = np.histogram2d(
                pos2_d[:, 0], pos2_d[:, 1], bins=(bins_x, bins_y), normed=True
            )
            correlation = np.mean(hist2 * hist1)
            if correlation > best_correlation:
                best_correlation = correlation
                best = np.array([dx, dy])
    
    return best
    """


def correct_drift(
    pos,
    L=0.2,
    step_size=0.03,
    prog_bar_position=None,
    min_n_locs_per_bin: int = 10000,
    max_n_bins: int = 20,
):
    pos["n_detection"] = np.arange(pos.shape[0])

    if "shift_x" not in pos.columns:
        pos[["shift_x", "shift_y"]] = np.zeros((pos.shape[0], 2))

    bin_size = max(min_n_locs_per_bin, int(pos.shape[0] / max_n_bins))
    bins = np.arange(0, pos.n_detection.max() + 1, bin_size)
    # print("Cut in %d bins" % len(bins))

    if bin_size > pos.shape[0] / 2:
        logging.info("Too few detections to correct drift")
        interp = None
    else:
        pos["n_bin"] = np.digitize(pos.n_detection, bins)
        sorted_bins = sorted(np.unique(pos.n_bin))
        n_steps = np.array(
            [pos.loc[pos.n_bin == b, "n_detection"].min() for b in sorted_bins]
        )

        pos_1 = [
            pos.loc[pos.n_bin == b, ["x", "y"]].copy().values
            for b in sorted_bins[:-1]
        ]
        pos_2 = [
            pos.loc[pos.n_bin == b, ["x", "y"]].copy().values
            for b in sorted_bins[1:]
        ]

        shifts_del = []
        for p1, p2 in tqdm(
            zip(pos_1, pos_2),
            leave=False,
            unit="bins",
            total=len(pos_1),
            position=0 if prog_bar_position is None else prog_bar_position,
            disable=False,  # Disable parce que compliqué en pooling
        ):
            shifts_del.append(
                da.from_delayed(
                    delayed(get_optimal_shift)(p1, p2, L=L, step=step_size),
                    dtype=float,
                    shape=(2,),
                )
            )
        shifts_del = da.stack(shifts_del, axis=0)
        shifts_del = da.concatenate([da.zeros((1, 2)), shifts_del], axis=0)
        shifts_del = da.cumsum(shifts_del, axis=0)
        shifts = shifts_del.compute()

        interp = interp1d(
            n_steps,
            shifts,
            axis=0,
            bounds_error=False,
            kind="quadratic" if pos.n_bin.nunique() > 2 else "linear",
            fill_value=(shifts[0], shifts[-1]),
        )
        interp_values = interp(pos.n_detection.values)
        pos[["x", "y"]] -= interp_values
        pos[["shift_x", "shift_y"]] += interp_values

    frames = np.arange(
        start=pos.n_detection.min(), stop=pos.n_detection.max(), step=1000
    )
    interp_df = pd.DataFrame(index=np.arange(len(frames)))
    interp_df["frame"] = frames
    interp_df["dx"] = 0.0
    interp_df["dy"] = 0.0

    if interp is not None:
        interp_df[["dx", "dy"]] = interp(interp_df["frame"])

    return pos, interp_df



from __future__ import annotations
import logging
from typing import Dict, List, Type, Union
import os
import pandas as pd
import pandas as pd
import dask.array as da
import yaml

from ..data_structure.acquisition import Acquisition
from ..data_structure.experiment import Experiment
from .steps import *


class TifPipeline:
    def __init__(
        self,
        name: str,
        movie_preprocessors: List[MoviePreProcessor],
        detector: Detector,
        localizer: SubpixelLocalizer,
        loc_processors: List[LocProcessor],
        tracker: Tracker,
    ):
        self.name = name
        self.movie_preprocessors = movie_preprocessors
        self.detector = detector
        self.localizer = localizer
        self.loc_processors = loc_processors
        self.tracker = tracker
        self.last_storage_path = None

    @classmethod
    def from_dict(cls, p: Dict):
        """
        Instantiate from a dictionnary.
        Here's an example Dictionnary which could be passed as an argument :

        .. code-block:: python3

            {
                "name":"my_pipeline",
                "movie_preprocessors":[
                    {
                        "MyPreProcessingClass":{"param1":value,"param2":other_value}
                    },
                    {
                        "WindowPercentileFilter":{}
                        # If the parameter's dict is empty, default parameters will be used
                    }
                }],
                "localizer":{
                    "DefaultLocalizer":{"threshold_factor":1.5}
                    },
                "tracker":{
                    "UnknownClass":{"bla":bla}
                    # If the class is not found, this will raise an exception
                    # Similarly, if the class provided does not inherit Tracker, an exception will be raised
                }
                # If some step is not mentioned (e.g. here, there's nothing about localization processing), then
                # if it's movie_preprocessors, then no movie_preprocessors will be used (same for loc_processors)
                # it it's localizer or tracker, then the default classes will be used.
            }

        """

        def instantiate_from_dict(d):
            assert len(d) <= 1, "Only one key can be provided"
            if len(d) == 0:
                return None
            cls = list(d.keys())[0]
            params = d[cls]
            klass = globals()[cls]
            instance = klass(**params)
            return instance

        def load_from_dict(d, key, default_class):
            instance = None
            if key in d:
                try:
                    instance = instantiate_from_dict(d[key])
                except Exception as e:
                    logging.debug(e)
                    raise
            if instance is None:
                instance = default_class()
            return instance

        def load_list_from_dict(d, key, base_class):
            instances = []
            if key in d:
                items_list = d[key]
                for item in items_list:
                    instance = instantiate_from_dict(item)
                    try:
                        assert isinstance(instance, base_class)
                        instances.append(instance)
                    except AssertionError as e:
                        logging.debug(e)
            return instances

        # LOAD COMPONENT BY COMPONENT
        name = p["name"]
        movie_preprocessors = load_list_from_dict(
            p, "movie_preprocessors", MoviePreProcessor
        )
        detector = load_from_dict(p, "detector", BaseDetector)
        localizer = load_from_dict(p, "localizer", MaxLikelihoodLocalizer)
        loc_processors = load_list_from_dict(p, "loc_processors", LocProcessor)
        tracker = load_from_dict(p, "tracker", ConservativeTracker)
        return cls(
            name=name,
            movie_preprocessors=movie_preprocessors,
            detector=detector,
            localizer=localizer,
            loc_processors=loc_processors,
            tracker=tracker,
        )

    @classmethod
    def from_yaml(cls, file):
        with open(file, "r") as f:
            params = dict(yaml.safe_load(f))
        res = cls.from_dict(params)
        res.last_storage_path = str(file)
        return res

    @classmethod
    def default_with_name(cls, name: str):
        return cls.from_dict({"name": name})

    @property
    def available_steps(self):
        if not hasattr(self, "_available_steps"):
            all_globals = globals().items()
            base_classes = {
                "movie_preprocessors": MoviePreProcessor,
                "detector": Detector,
                "localizer": SubpixelLocalizer,
                "loc_processors": LocProcessor,
                "tracker": Tracker,
            }
            self._available_steps = {}
            for step_type in base_classes:
                self._available_steps[step_type] = []

            for name, obj in all_globals:
                try:
                    for step_type, base_class in base_classes.items():
                        if (
                            issubclass(obj, base_class)
                            and obj is not base_class
                        ):
                            self._available_steps[step_type].append(
                                (name, obj)
                            )
                            break
                except TypeError as e:
                    pass

        return self._available_steps

    def to_yaml(self, fileName):
        tp_params = self.to_dict()
        yaml.dump(tp_params, open(fileName, "w"))
        self.last_storage_path = str(fileName)

    def to_dict(self) -> dict:
        res = {
            "name": self.name,
            "detector": self.detector.to_dict(),
            "localizer": self.localizer.to_dict(),
            "tracker": self.tracker.to_dict(),
        }
        if len(self.movie_preprocessors) > 0:
            res["movie_preprocessors"] = [
                p.to_dict() for p in self.movie_preprocessors
            ]
        if len(self.loc_processors) > 0:
            res["loc_processors"] = [p.to_dict() for p in self.loc_processors]
        return res

    def contains_class(self, step):
        return self.index_of(step) is not None

    def step_type_of(self, step):
        d = self.available_steps
        for step_type, steps in d.items():
            for step_tuple in steps:
                if step_tuple[0] == step:
                    return step_type
        return None

    def step_class_of(self, step):
        d = self.available_steps
        for step_type, steps in d.items():
            for step_tuple in steps:
                if step_tuple[0] == step:
                    return step_tuple[1]
        return None

    def index_of(self, step):
        d = self.to_dict()
        for step_type, steps in d.items():
            if step_type == "name":
                continue
            if isinstance(steps, list):
                # Si possible d'avoir plusieurs items, on trouve le rang
                for step_dict in steps:
                    for i, option in enumerate(step_dict):
                        if step == option:
                            return step_type
            else:
                assert isinstance(steps, dict), steps
                # Sinon, on renvoie zéro parce qu'il n'y a forcément qu'un step
                if step in steps:
                    return 0
        return None

    def can_be_removed(self, step):
        return self.contains_class(step) and (
            self.step_type_of(step)
            in ["movie_preprocessors", "loc_processors"]
        )

    def is_mandatory(self, step):
        return self.step_type_of(step) not in [
            "movie_preprocessors",
            "loc_processors",
        ]

    def has_alternatives_to(self, step):
        step_type = self.step_type_of(step)
        if step_type is None:
            return False
        else:
            return len(self.available_steps[step_type]) > 1

    def __str__(self):
        desc = "TIF Processing pipeline\n"
        desc += "-----------------------\n"
        desc += "Movie preprocessing steps :\n"
        for i, step in enumerate(self.movie_preprocessors):
            desc += "%d/%d \t %s\n" % (
                i + 1,
                len(self.movie_preprocessors),
                step,
            )
        desc += "-----------------------\n"
        desc += "Detector :\n"
        desc += "\t %s\n" % self.detector
        desc += "-----------------------\n"
        desc += "Localizer :\n"
        desc += "\t %s\n" % self.localizer
        desc += "-----------------------\n"
        desc += "Localization processing steps :\n"
        for i, step in enumerate(self.loc_processors):
            desc += "- %d/%d \t %s\n" % (i + 1, len(self.loc_processors), step)
        desc += "-----------------------\n"
        desc += "Tracker :\n"
        desc += "\t %s\n" % self.tracker
        return desc

    def movie_preprocessing(self, mov: da.Array) -> da.Array:
        for i, processor in enumerate(self.movie_preprocessors):
            logging.info(
                "Preprocessing step %d / %d : %s"
                % (i, len(self.movie_preprocessors), processor.name)
            )
            mov = processor.preprocess(mov)
        return mov

    def movie_localization(
        self, mov: da.Array, DT: float, pixel_size: float
    ) -> pd.DataFrame:
        detections = self.detector.movie_detection(mov)
        locs = self.localizer.movie_localization(mov, detections)
        locs[["x", "y"]] *= pixel_size
        locs["t"] = locs["frame"] * DT
        return locs

    def loc_processing(
        self,
        mov: da.Array,  # Movie
        locs: pd.DataFrame,  # Previous localizations, with x and y in micrometers
        pixel_size: float = 1.0,  # Pixel size in micrometers
    ) -> pd.DataFrame:
        for i, locproc in enumerate(self.loc_processors):
            logging.info(
                "Processing locs, step %d / %d : %s"
                % (i, len(self.loc_processors), locproc.name)
            )
            locs = locproc.process(mov, locs, pixel_size=pixel_size)
        return locs

    def tracking(self, locs: pd.DataFrame) -> pd.DataFrame:
        return self.tracker.track(locs)

    def _process(
        self,
        acq: Acquisition,
        skip_loc: bool = False,
        skip_tracking: bool = False,
    ):
        # Load the original .tif
        if skip_tracking and not skip_loc:
            skip_tracking = False
            logging.warning(
                "Can't skip tracking after fresh localization. Overriding to skip_tracking = False"
            )
        original_mov = acq.image.astype(float)
        if not skip_loc:
            mov = self.movie_preprocessing(original_mov)
            locs = self.movie_localization(
                mov, DT=acq.experiment.DT, pixel_size=acq.experiment.pixel_size
            )
            acq.raw_locs = locs.copy()
            self.mark_as_localized(acq)
        if not skip_tracking:
            acq.experiment.pixel_size
            locs = self.loc_processing(original_mov, acq.raw_locs.copy())
            acq.locs = self.tracking(locs)
            self.mark_as_tracked(acq)

    def process(
        self,
        to_process: Union[Acquisition, Experiment],
        force_reprocess: bool = False,
    ):
        logging.debug("Process %s" % to_process)
        if isinstance(to_process, Acquisition):
            # Actually process
            self._process(
                to_process,
                skip_loc=(not force_reprocess)
                and self.is_already_localized(to_process),
                skip_tracking=(not force_reprocess)
                and self.is_already_tracked(to_process),
            )
        elif isinstance(to_process, Experiment):

            # Actually process, acquisition after acquisition

            for i, acq_file in enumerate(to_process):
                acq = Acquisition(
                    acq_file, experiment=to_process, tif_pipeline=self
                )
                if i == 0:
                    pipeline_export_path_for_exp = self.exp_params_path(acq)
                    logging.info(
                        "Saving pipeline to %s" % pipeline_export_path_for_exp
                    )
                    self.to_yaml(pipeline_export_path_for_exp)
                self.process(acq, force_reprocess=force_reprocess)
        else:
            raise BaseException(
                "to_process shoud be an experiment or an acquisition but is %s"
                % to_process.__class__.__qualname__
            )

    def exp_run_df_path(self, acq: Acquisition) -> str:
        run_path = os.path.join(acq.experiment.export_folder, self.name)
        if not os.path.exists(run_path):
            os.mkdir(run_path)
        assert os.path.isdir(run_path)
        exp_run_df_path = os.path.join(
            acq.experiment.export_folder, self.name, "run_index.csv"
        )
        return exp_run_df_path

    def exp_params_path(self, acq: Acquisition) -> str:
        run_path = os.path.join(acq.experiment.export_folder, self.name)
        if not os.path.exists(run_path):
            os.mkdir(run_path)
        assert os.path.isdir(run_path)
        exp_params_path = os.path.join(
            acq.experiment.export_folder, self.name, "%s.yaml" % self.name
        )
        return exp_params_path

    def is_already_localized(self, acq: Acquisition):
        p = self.exp_run_df_path(acq=acq)
        if os.path.exists(p):
            run_index_df = pd.read_csv(p, index_col=0)
            try:
                return run_index_df.loc[acq.ID, "is_localized"] == True
            except:
                pass
        return False

    def is_already_tracked(self, acq: Acquisition):
        p = self.exp_run_df_path(acq=acq)
        if os.path.exists(p):
            run_index_df = pd.read_csv(p, index_col=0)
            try:
                return run_index_df.loc[acq.ID, "is_tracked"] == True
            except:
                pass
        return False

    def mark_as_localized(self, acq: Acquisition):
        p = self.exp_run_df_path(acq)
        if os.path.exists(p):
            df = pd.read_csv(p, index_col=0)
        else:
            df = pd.DataFrame(
                columns=["is_localized", "is_tracked"], dtype=bool
            )
        df.loc[acq.ID, "is_localized"] = True
        df.to_csv(p, index=True)

    def mark_as_tracked(self, acq: Acquisition):
        p = self.exp_run_df_path(acq)
        if os.path.exists(p):
            df = pd.read_csv(p, index_col=0)
        else:
            df = pd.DataFrame(
                columns=["is_localized", "is_tracked"], dtype=bool
            )
        df.loc[acq.ID, "is_tracked"] = True
        df.to_csv(p, index=True)



__version__ = "0.0.1"


from ._widget import PalmariWidget
from .data_structure.experiment import Experiment
from .data_structure.acquisition import Acquisition
from .processing.tif_pipeline import TifPipeline

# from .processing.tif_pipeline_widget import TifPipelineWidget



from __future__ import annotations
import json
import logging
from glob import glob
import secrets
from subprocess import call
import pandas as pd
import os
import json
import matplotlib.pyplot as plt
from typing import List
import numpy as np
from typing import TYPE_CHECKING

from .acquisition import Acquisition


class Experiment:
    def __init__(
        self,
        data_folder: str,
        export_folder: str,
        DT: float = None,
        pixel_size: float = None,
        file_pattern: str = None,
    ):
        self.data_folder = data_folder
        self.export_folder = export_folder
        self.DT = DT
        self.pixel_size = pixel_size
        self.file_pattern = file_pattern
        self.check_export_folder_and_load_info()
        self.index_path = os.path.join(self.export_folder, "index.csv")
        self.look_for_updates()
        logging.info(self)

    def __str__(self):
        desc = "Data : %s\n" % self.data_folder
        desc += "Export : %s\n" % self.export_folder
        desc += "Files : %d\n" % self.index_df.shape[0]
        n_cols = len(self.index_df.columns) - 2
        desc += "Information columns : %d\n" % n_cols
        for col in self.index_df.columns:
            if col not in ["file", "ID"]:
                desc += "\t- %s\n" % (col)
        desc += "Exposure : %.3f second(s)\n" % self.DT
        desc += "Pixel size : %.3f micron(s)\n" % self.pixel_size
        return desc

    def __iter__(self):
        return iter(self.index_df.file.tolist())

    def __getitem__(self, item):
        if item in self.index_df.ID.tolist():
            file = self.index_df.loc[self.index_df.ID == item, "file"].values[
                0
            ]
            return file
        elif type(item) == int:
            if item < self.index_df.shape[0]:
                return self.index_df["file"].values[item]
        else:
            raise "Invalid item %s" % item

    def __len__(self):
        return self.index_df.shape[0]

    @classmethod
    def from_single_tif(cls, tif_file: str, export_folder: str):
        data_folder = os.path.split(tif_file)[0]
        return Experiment(data_folder=data_folder, export_folder=export_folder)

    """
    def to_tracksets(self, tif_pipeline: TifPipeline) -> TrackSets:
        tracksets = []
        included_files = []
        for tif_file in self:
            acq = Acquisition(
                tif_file=tif_file, experiment=self, tif_pipeline=tif_pipeline
            )
            if not acq.is_processed:
                logging.debug("Not including %s" % tif_file)
                continue
            included_files.append(tif_file)
            tracksets.append(
                TrackSet(locs_path=acq.locs_path, origin_file=tif_file)
            )
        return TrackSets(
            tracksets=tracksets,
            root_folder=self.export_folder,
            index_df=self.index_df.loc[
                self.index_df.file.isin(included_files)
            ],
        )
    """

    def check_export_folder_and_load_info(self):
        json_path = os.path.join(self.export_folder, "exp_info.json")
        if not os.path.isdir(self.export_folder):
            # Create a json file with experiment-level information
            logging.info("Create %s" % self.export_folder)
            os.mkdir(self.export_folder)

        if not os.path.exists(json_path):
            logging.info("JSON does not exist, create %s" % json_path)
            # Quite some redundancies here...
            pixel_size = self.pixel_size
            if pixel_size is None:
                pixel_size = 0.097
                try:
                    pixel_size = float(
                        input("Pixel size, in microns (default is 0.097) : ")
                    )
                except BaseException as e:
                    print(e)
                    print(
                        "Incorrect value, keeping default : %.2f um"
                        % pixel_size
                    )
                self.pixel_size = pixel_size

            DT = self.DT
            if DT is None:
                DT = 0.03
                try:
                    DT = float(
                        input(
                            "Time interval between successive frames, in seconds (default is 0.03) : "
                        )
                    )
                except BaseException as e:
                    print(e)
                    print("Incorrect value, keeping default : %.2f s" % DT)
                self.DT = DT

            file_pattern = self.file_pattern
            if file_pattern is None:
                file_pattern = ".tif"
                try:
                    input_file_pattern = str(
                        input('File pattern (default is ".tif") : ')
                    )
                    if len(input_file_pattern) > 0:
                        file_pattern = input_file_pattern

                except BaseException as e:
                    print(e)
                    print("Incorrect value, keeping default : %.2f s" % DT)
                self.file_pattern = file_pattern

            exp_info = {
                "data_path": self.data_folder,
                "pixel_size": self.pixel_size,
                "file_pattern": self.file_pattern,
                "DT": self.DT,
            }
            json.dump(
                exp_info,
                open(json_path, "w"),
            )
        else:
            # Read the JSON file and check that path of origin data is the same
            logging.info("JSON already exists : %s" % json_path)
            exp_info = json.load(open(json_path, "r"))

            if not os.path.samefile(self.data_folder, exp_info["data_path"]):
                logging.debug("Data folder is %s" % self.data_folder)
                logging.debug(
                    "Data path in JSON in %s" % exp_info["data_path"]
                )
        self.DT = exp_info["DT"]
        self.pixel_size = exp_info["pixel_size"]
        self.file_pattern = exp_info["file_pattern"]

    @property
    def index_df(self) -> pd.DataFrame:
        if not hasattr(self, "_index_df"):
            if not os.path.exists(self.index_path):
                self._index_df = pd.DataFrame(columns=["file", "ID"])
            else:
                self._index_df = pd.read_csv(self.index_path)
        return self._index_df

    @index_df.setter
    def index_df(self, v: pd.DataFrame):
        self._index_df = v
        self.save_index()

    def save_index(self):
        """
        Just saves index_df
        """
        self.index_df.to_csv(self.index_path, index=False)

    @property
    def custom_fields(self) -> dict:
        """
        Override this in a subclass of Experience to meet your needs
        keys of the dict are column names
        values are used to fill the columns, using the TIF file name of each acquisition
        values can be :

        - string : True if the file name contains that string
        - int : the i-th part of the file name, when split using the filesystem separator
        - callable : callable(filename)

        for instance

        .. code-block:: python3

            {"condition":get_condition_from_name}

        """
        return {}

    """
    @property
    def available_processing_runs(self) -> List[TifProcessingRun]:
        if not hasattr(self, "_processing_runs"):
            os.chdir(self.export_folder)
            run_param_files = glob("*/run_params.json", recursive=True)
            logging.info("%d available processing runs" % len(run_param_files))
            self._processing_runs = [
                TifProcessingRun.from_json(f, self) for f in run_param_files
            ]
        return self._processing_runs
    """

    @property
    def all_files(self) -> List[str]:
        """Return all files indexed in self.index_df

        Returns:
            List[Acquisition]: all files indexed in self.index_df
        """
        if not hasattr(self, "_all_files"):
            # Call it with no filters to retrieve all files
            self._all_files = self.files_with_filters()
            logging.info(
                "Querying all files : Found %d files" % len(self._all_files)
            )
            if len(self._all_files) > 0:
                logging.info(
                    "The first queried file is %s" % self._all_files[0]
                )
        return self._all_files

    """
    def files_with_filters(
        self, filters: dict = {}, only_processed_with_run: TifProcessingRun = None
    ) -> List[str]:
        # filters = {"cell_type":["neuron","platelet"],
        #            "replicate":"Experience0"}
        cond = ~self.index_df.file.isnull()
        for column, value in filters.items():
            try:
                if isinstance(value, list):
                    cond = cond & self.index_df[column].isin(value)
                else:
                    cond = cond & (self.index_df[column] == value)
            except KeyError:
                logging.debug("%s is not a column" % column)
                logging.debug("Use one of %s" % ", ".join(self.index_df.columns))
                raise
        files = self.index_df.loc[cond].file.tolist()
        if only_processed_with_run is not None:
            files = [
                f
                for f in files
                if Acquisition(
                    f, experiment=self, tif_pipeline=only_processed_with_run
                ).is_processed
            ]
        return files
    """

    def add_new_roi_to_index(self, f: str):
        row = pd.DataFrame.from_dict(
            {"file": [f], "ID": [secrets.token_hex(8)]}, orient="columns"
        )
        self.index_df = pd.concat(
            [self.index_df, row], axis=0, ignore_index=True
        )
        logging.info("Added row for %s" % f)
        logging.info(str(row))

    def remove_old_roi_from_index(self, f: str):
        assert f in self.index_df["file"].tolist()
        self.index_df = self.index_df.loc[self.index_df.file != f]
        logging.info("Removed in index the row %s" % f)

    def scan_folder(self) -> list:
        os.chdir(self.data_folder)
        roi_files = glob("**/*%s" % self.file_pattern, recursive=True)
        roi_files = [
            f for f in roi_files if os.path.getsize(f) > 1e6
        ]  # Don't consider files of less than 1Mb
        # TODO: improve filtering here
        return roi_files

    def look_for_updates(self):
        """
        Look if the index dataframe matches
        the reality of present/absent files
        And computes custom columns if needed
        """
        # D'abord, on vérifie qu'on a une ligne par fichier .tif
        roi_files = self.scan_folder()
        logging.info("look for updates : found %d files" % len(roi_files))
        for f in roi_files:
            if f not in self.index_df["file"].tolist():
                self.add_new_roi_to_index(f)

        # Ensuite on vérifie que tous les fichiers référencés dans l'index
        # sont encore là
        if self.index_df.shape[0] == 0:
            return
        for f in self.index_df["file"].tolist():
            if f not in roi_files:
                self.remove_old_roi_from_index(f)

        self.look_for_new_columns()

    def get_ID_of_acq(self, acquisition: Acquisition):
        ID = self.index_df.loc[
            self.index_df.file == acquisition.tif_file, "ID"
        ].values[0]
        logging.debug("ID of file %s is %s" % (acquisition.tif_file, ID))
        return str(ID)

    def look_for_new_columns(self, overwrite=False):
        """
        Computes custom columns

        Args:
            overwrite (bool, optional): Whether to overwrite pre-existing values. Defaults to False.
        """
        for column, value in self.custom_fields.items():

            if callable(value):
                values = self.index_df["file"].apply(value)
            elif type(value) is int:
                try:
                    s = self.index_df["file"].str.split(
                        pat=os.path.sep, expand=True
                    )
                    i = value
                    if i < 0:
                        i = s.shape[1] + i
                    values = s[i]
                except KeyError as e:
                    logging.debug(
                        self.index_df["file"].str.split(
                            pat=os.path.sep, expand=True
                        )
                    )
                    logging.debug(e)
                    raise
            elif type(value) is str:

                def last_part_that_contains(f):
                    parts = f.lower().split(os.path.sep)
                    for p in parts[::-1]:
                        if value in p:
                            return p

                values = self.index_df["file"].map(last_part_that_contains)
            else:
                values = pd.Series(index=self.index_df.index)
                values[:] = value
            if column in self.index_df.columns and not overwrite:
                values.loc[
                    ~self.index_df[column].isnull()
                ] = self.index_df.loc[~self.index_df[column].isnull(), column]
            self.index_df[column] = values

    # AGGREGATE STATS

    @property
    def runs_stats(self) -> pd.DataFrame:
        if not hasattr(self, "_runs_stats"):
            runs = self.available_processing_runs
            logging.info(
                "Concatenating stats from %d sets of parameters" % len(runs)
            )
            self._runs_stats = pd.concat(
                [run.stats_df for run in runs], axis=0
            )

        return self._runs_stats

    def parameter_influence_on_stats(
        self, param_name: str, stats: list = ["n_locs"], mode: str = "hist"
    ):
        """Plots the influence of a processing parameter on some statistics.

        Args:
            param_name (str): The parameter whose influence is studied. it should be one returned by acquisition.basic_stats
            stats (list, optional): Statistics to compare. Defaults to ["n_locs"].
            mode (str, optional): Plotting mode. supported : "bars" and "hist". Defaults to "hist".
        """
        stat_cols = [c for c in self.runs_stats.columns if c != "file"]
        df = self.runs_stats
        pv = df.pivot_table(
            columns=param_name,
            index="file",
            values=stat_cols,
            aggfunc="median",
        )
        param_values = df[param_name].unique().tolist()

        if mode == "bars":
            w = 1.0 / (1 + len(stats))
            for s in stats:
                fig = plt.figure()
                ax = fig.add_subplot(111)
                for i, v in enumerate(param_values):
                    label = None
                    if type(v) is float:
                        label = "%s = %.2f" % (param_name, v)
                    else:
                        label = "%s = %s" % (param_name, v)
                    ax.bar(
                        np.arange(pv.shape[0])
                        - w * (len(stats) - 1) / 2
                        + i * w,
                        height=pv[(s, v)],
                        width=w,
                        tick_label=pv.index,
                        label=label,
                    )
                if self.runs_stats[s].max() > 100 * self.runs_stats[s].min():
                    plt.yscale("log")
        elif mode == "hist":
            for s in stats:
                fig = plt.figure(figsize=(4, 2))
                ax = fig.add_subplot(111)
                log = self.runs_stats[s].max() > 100 * self.runs_stats[s].min()
                title = s
                if log:
                    title = "log_10(%s)" % title
                ax.set_title(title)
                for i, v in enumerate(param_values):
                    label = None
                    if type(v) is float:
                        label = "%s = %.2f" % (param_name, v)
                    else:
                        label = "%s = %s" % (param_name, v)
                    H = pv[(s, v)]
                    if log:
                        H = np.log10(1 + H.astype(float).values)
                    ax.hist(H, bins=20, label=label, histtype="step")
                ax.legend()



from .base import *


class ConservativeTracker(Tracker):
    def __init__(self, max_diffusivity: float = 5.0):
        # Attributes will automatically be detected as parameters of the step and stored/loaded.
        # Parameters must have default values
        self.max_diffusivity = max_diffusivity

    def track(self, locs: pd.DataFrame):
        # This is where the actual tracking happen.
        import trackpy as tp

        delta_t = self.estimate_delta_t(locs)  # This is a Tracker's method.
        dim = 2
        max_radius = np.sqrt(2 * dim * self.max_diffusivity * delta_t)
        logging.info("Max radius is %.2f" % max_radius)
        tracks = tp.link(locs, search_range=max_radius, link_strategy="drop")
        locs["n"] = tracks["particle"]
        return locs

    @property
    def name(self):
        # This is for printing
        return "Default tracker (Trackpy)"

    # The following dicts are used when setting the parameters through a graphic interface, using open_in_napari()
    widget_types = {
        "max_diffusivity": "FloatSpinBox",
        "delta_t": "FloatSpinBox",
    }
    # For details about widget types, see https://napari.org/magicgui/
    widget_options = {
        "delta_t": {
            "step": 0.01,
            "tooltip": "time interval between frames (in seconds)",
            "min": 0.0,
            "label": "Time delta (s)",
        },
        "max_diffusivity": {
            "step": 1.0,
            "tooltip": "Assumed maximum diffusivity (in microns per square second).\nThis is used in conjunction with the Time delta to set the maximal distance between consecutive localizations",
            "label": "D_max (um^2/s)",
            "min": 0.0,
        },
    }



from .base import *
from ...tif_tools.correct_drift import correct_drift
from dask_image.ndfilters import gaussian_filter
from .quot_localizer import BaseDetector, MaxLikelihoodLocalizer
from scipy.ndimage import gaussian_filter1d
from scipy.interpolate import interp1d


class CorrelationDriftCorrector(LocProcessor):

    widget_type = {"max_n_bins": "Spinner", "min_n_locs_per_bin": "Spinner"}
    widget_options = {
        "max_n_bins": {
            "min": 2,
            "max": 50,
            "label": "Bins",
            "tooltip": "Maximum number of bins. One shift per bin will be computed.",
        },
        "min_n_locs_per_bin": {
            "min": 100,
            "max": int(1e7),
            "step": 1000,
            "label": "Locs / bin",
            "tooltip": "Minimum number of localizations per bin",
        },
    }

    def __init__(self, min_n_locs_per_bin: int = 10000, max_n_bins: int = 20):
        self.min_n_locs_per_bin = min_n_locs_per_bin
        self.max_n_bins = max_n_bins

    def process(
        self, mov: da.Array, locs: pd.DataFrame, pixel_size: float
    ) -> pd.DataFrame:

        logging.info("Correcting drift iteration #1/2")
        df, shift_1 = correct_drift(
            locs,
            L=0.2,
            step_size=0.03,
            min_n_locs_per_bin=self.min_n_locs_per_bin,
            max_n_bins=self.max_n_bins,
        )
        logging.info("Correcting drift iteration #2/2")
        df, shift_2 = correct_drift(
            df,
            L=0.05,
            step_size=0.01,
            min_n_locs_per_bin=self.min_n_locs_per_bin,
            max_n_bins=self.max_n_bins,
        )

        # this causes a crash
        # plt.figure()
        # plt.plot(df["frame"], df["shift_x"], label="x shift")
        # plt.plot(df["frame"], df["shift_y"], label="y shift")
        # plt.legend()
        # plt.show()
        return df

    @property
    def name(self):
        return "Correlation-based drift corrector"

    @property
    def action_name(self):
        return "Correct drift"


class BeadDriftCorrector(LocProcessor):

    widget_types = {"sigma": "FloatSpinBox"}
    widget_options = {
        "sigma": {
            "min": 2,
            "max": 50,
            "label": "Bead radius (in pixels)",
        },
    }

    def __init__(self, sigma: int = 3):
        self.sigma = sigma
        self._n_frames = (
            100  # Width of the Gaussian kernel along the 'frame' axis
        )

    def process(
        self, mov: da.Array, locs: pd.DataFrame, pixel_size: float
    ) -> pd.DataFrame:

        k = 10
        threshold_mov = (mov / mov.max(axis=(1, 2), keepdims=True)).astype(
            float
        )
        gaussian_mov = gaussian_filter(
            threshold_mov,
            sigma=(self._n_frames, self.sigma, self.sigma),
            mode="nearest",
            truncate=3,
        )[::k]
        detector = BaseDetector(
            k=self.sigma, w=int(10 * self.sigma) + 1, t=10.0
        )
        detections = detector.process(gaussian_mov)

        localizer = MaxLikelihoodLocalizer(
            window_size=int(6 * self.sigma) + 1, sigma=self.sigma
        )
        _pos = localizer.process(gaussian_mov, detections)
        bead_pos = (
            (_pos.sort_values("I0", ascending=False).groupby("frame").first())
            .reset_index()
            .sort_values("frame")
        )
        bead_pos["frame"] = (bead_pos["frame"] - 1) * k + 1

        shift = bead_pos[["x", "y"]].values * pixel_size

        # print("Shift std :")
        # print(_pos[["x", "y"]].std(axis=0))
        # print("Locs std : ")
        # print(locs[["x", "y"]].std(axis=0))
        shift = shift - shift[0]
        t = bead_pos["frame"].values
        plt.figure()
        plt.plot(t, shift[:, 0], label="shift")

        f = interp1d(t, shift, kind="linear", axis=0, fill_value="extrapolate")
        computed_shift = f(locs["frame"])

        plt.plot(locs["frame"], computed_shift[:, 0], label="computed")
        plt.legend()

        locs[["x", "y"]] -= computed_shift
        return locs

    @property
    def name(self):
        return "Bead-based drift corrector"

    @property
    def action_name(self):
        return "Correct drift with bead"



#!/usr/bin/env python
"""
quot.filters -- background subtraction and related filters
to clean up dirty data for detection and localization

"""
# Numeric 
import numpy as np 

# Gaussian filtering
from scipy.ndimage import gaussian_filter 

# File reader
from .read import ImageReader

class ChunkFilter(ImageReader):
    """
    A chunkwise file reader that also performs common filtering
    operations, like subtracting background from each frame.

    It is convenient to integrate file reading and filtering, since
    filtering often involves computing some kind of function on
    a chunk of frames. For instance, we can perform a rudimentary
    kind of background subtraction by subtracting the minimum value
    of each pixel in a 100-frame block from the corresponding pixel
    in each individual frame.

    When iterating over this object, the ChunkFilter does the
    following:

        1. Load a new chunk, cached at self.chunk.
        2. Perform some precomputations on the chunk - for instance,
            find the through-time minimum of each pixel in frame.
        3. Return individual filtered frames using the calculated
            precomputations.
        4. Load new chunks as necessary.

    init
        path            :   str, path to an image file
        start           :   int, the start frame when iterating
        stop            :   int, the stop frame when iterating
        method          :   str, the filtering method to use
        chunk_size      :   int, the block size to use for filtering
        method_static   :   bool, the method is unlikely to change
                            during the existence of this ChunkFilter.
                            This is not true, for instance, when using
                            the ChunkFilter in a GUI setting where the
                            user can rapidly change between different 
                            filtering settings. If False, the ChunkFilter
                            caches information about the chunk much more
                            aggressively. This makes switching between 
                            filtering methods much faster at the cost of 
                            memory.
        init_load       :   bool, load the first chunk during init.
                            Generally keep this True unless you have a 
                            very good reason not to.
        **method_kwargs :   to the filtering method

    attributes
        self.cache
        self.block_starts (1D ndarray, int)
        self.block_start (int), the first frame for the current
            block

    """
    def __init__(self, path, start=None, stop=None, method='identity',
        chunk_size=40, method_static=True, init_load=True, **method_kwargs):

        # Open file reader
        super(ChunkFilter, self).__init__(path, start=start, stop=stop)

        # If the filtering method is unset, default to the 
        # raw image
        if method is None:
            method = 'identity'

        self.method = method 
        self.chunk_size = chunk_size 
        self.method_static = method_static 
        self.method_kwargs = method_kwargs 

        # If the chunk size is larger than the number of frames,
        # set it to the number of frames
        if self.chunk_size > self.n_frames:
            self.chunk_size = self.n_frames

        # Initially, set no subregion on the image.
        self.sub_kwargs = {}

        # Assign each frame index to a block
        self._init_chunks()

        # Set the initial chunk
        self.chunk_start = self.get_chunk_start(self.start)

        # Load the initial chunk 
        if init_load:
            self.load_chunk(self.chunk_start)

        # Set the method keyword arguments
        self.set_method_kwargs(method=self.method, **method_kwargs)

    def __next__(self):
        if self._c < self.stop:
            self._c += 1
            return self.filter_frame(self._c-1)
        else:
            raise StopIteration 

    def _init_chunks(self):
        """
        Choose which frames correspond to which chunks by instantiating
        an array with the starting frame index of each chunk.

        """
        n_chunks = self.n_frames//self.chunk_size + 1
        self.chunk_starts = np.zeros(n_chunks+1, dtype='int64')
        self._offset = self.start % self.chunk_size
        self.chunk_starts[1:] = np.arange(n_chunks)*self.chunk_size+self._offset
        self.chunk_starts = self.chunk_starts[self.chunk_starts<self.n_frames]

        # Special case: last chunk 
        self.chunk_starts[-1] = self.n_frames - self.chunk_size

    def _in_chunk(self, frame_index):
        """
        Return True if the frame is in the current chunk.

        """
        return (frame_index>=self.chunk_start) and \
            (frame_index<self.chunk_start+self.chunk_size)

    def _generate_cache(self):
        """
        Cache some factors calculated on the current chunk.

        If *method_static* is True, then only compute factors
        necessary for the current filtering method.

        Otherwise compute factors necessary for ANY filtering 
        method from the chunk. This is useful when the user is
        expected to change quickly between methods - for instance,
        in a GUI setting.

        """
        if not self.method_static:
            self.cache = {
                'mean_img': self.chunk.mean(axis=0),
                'median_img': np.median(self.chunk, axis=0),
                'min_img': self.chunk.min(axis=0),
            }
            for i in ['mean_img', 'median_img', 'min_img']:
                self.cache['gauss_filt_%s' % i] = gaussian_filter(self.cache[i], self.method_kwargs.get('k', 5.0))
        else:
            self.cache = {}
            if FILTER_CACHES[self.method] == 'mean_img':
                self.cache['mean_img'] = self.chunk.mean(axis=0)
            elif FILTER_CACHES[self.method] == 'median_img':
                self.cache['median_img'] = np.median(self.chunk, axis=0)
            elif FILTER_CACHES[self.method] == 'min_img':
                self.cache['min_img'] = self.chunk.min(axis=0)
            elif FILTER_CACHES[self.method] == 'gauss_filt_mean_img':
                self.cache['gauss_filt_mean_img'] = gaussian_filter(
                    self.chunk.mean(axis=0),
                    self.method_kwargs.get('k', 5.0)
                )
            elif FILTER_CACHES[self.method] == 'gauss_filt_min_img':
                self.cache['gauss_filt_min_img'] = gaussian_filter(
                    self.chunk.min(axis=0),
                    self.method_kwargs.get('k', 5.0)
                )               
            elif FILTER_CACHES[self.method] == 'gauss_filt_median_img':
                self.cache['gauss_filt_median_img'] = gaussian_filter(
                    np.median(self.chunk, axis=0),
                    self.method_kwargs.get('k', 5.0)
                )
            else:
                pass 

    def _recache_filtered_image(self):
        """
        A somewhat specialized method. Recache only one of the Gaussian-
        filtered images, like "gauss_filt_min_img", "gauss_filt_median_img",
        etc. This is useful when setting method kwargs that will change
        the kernel size for filtering.

        This is only permissible when self.method_static is False, since
        it requires using one of the basic images ("min_img" or "median_img",
        in the two examples above).

        """
        if self.method_static or (not (self.method in ['sub_gauss_filt_min',
            'sub_gauss_filt_median', 'sub_gauss_filt_mean'])):
            print("ChunkFilter._recache_filtered_image: has no effect " \
                "when method is static or does not involve Gaussian filtering")
        else:
            image_name = FILTER_CACHES[self.method]
            self.cache[image_name] = gaussian_filter(self.cache[ORIGIN_IMAGES[image_name]],
                self.method_kwargs.get('k', 5.0))
            self.method_kwargs[image_name] = self.cache[image_name]

    def set_method_kwargs(self, method=None, recache_filtered_image=False,
        **kwargs):
        """
        Set the kwargs passed to the underlying filtering 
        method. This includes factors that are precomputed
        for each chunk.

        If *method* is None, then use the current method.

        Sets the self.method_kwargs variable.

        args
        ----
            method  :   str, the name of the method to use
            recache :   bool, force the ChunkFilter to recache
                        the Gaussian filtered image, if it exists
            kwargs  :   to the method

        """
        self.method_kwargs = kwargs 

        # Regenerate the cache for a new method if necessary
        if not (method is None):
            self.method = method 
            if self.method_static:
                self._generate_cache()

        # Force re-cache, if desired
        if recache_filtered_image:
            self._recache_filtered_image()
        
        # Set cached factors required as keyword arguments
        if not (FILTER_CACHES[self.method] is None):
            self.method_kwargs[FILTER_CACHES[self.method]] = \
                self.cache[FILTER_CACHES[self.method]]

    def set_chunk_size(self, chunk_size):
        """
        Change the chunk size for this ChunkFilter.

        args
        ----
            chunk_size  :   int

        """
        self.chunk_size = chunk_size
        self._init_chunks()
        self.load_chunk(self.chunk_start)
        self._generate_cache()

    def set_subregion(self, **kwargs):
        """
        Set a subregion for this ChunkFilter. All filtering methods will
        return only this subregion.

        args
            y0, y1, x0, x1 (int), the limits of the subregion

        """
        self.sub_kwargs = self._process_subregion(**kwargs)
        self.chunk_height = self.sub_kwargs['y1'] - \
            self.sub_kwargs['y0']
        self.chunk_width = self.sub_kwargs['x1'] - \
            self.sub_kwargs['x0']

    def get_chunk_start(self, frame_index):
        """
        Return the starting frame for the corresponding chunk.

        """
        return self.chunk_starts[(frame_index-self._offset)//self.chunk_size+1]

    def load_chunk(self, start, recache=True):
        """
        Load a new image chunk starting at the frame *start*.
        Saves the chunk at self.chunk.

        args
            start       :   int
            recache     :   bool, recompute cached factors for the
                            new chunk

        """
        # If too close to the end of the movie, pick up 
        # frames before *start* to have a full chunk
        if start > (self.n_frames-self.chunk_size):
            self.local_index = start-self.n_frames+self.chunk_size
            start = self.n_frames-self.chunk_size 
        else:
            self.local_index = 0

        # Read the chunk from the file
        self.chunk = self.get_subregion_range(start=start,
            stop=start+self.chunk_size,
            **self.sub_kwargs).astype('float64')

        # Precompute some values for the current filtering method
        if recache:
            self._generate_cache()
            cache_key = FILTER_CACHES[self.method]
            if not cache_key is None:
                self.method_kwargs[cache_key] = self.cache[cache_key]

    def filter_frame(self, frame_index):
        """
        Return a single filtered frame from the movie.

        args
            frame_index     :   int

        returns
            2D ndarray (YX)

        """
        assert self._frame_valid(frame_index)

        if not self._in_chunk(frame_index):
            self.chunk_start = self.get_chunk_start(frame_index)
            self.load_chunk(self.chunk_start)

        return FILTER_METHODS.get(self.method)(
            self.chunk[frame_index-self.chunk_start], 
            **self.method_kwargs)

#######################
## FILTERING METHODS ##
#######################

def identity(img, **kwargs):
    """
    Do not filter the image.

    args
    ----
        img     :   2D ndarray (YX)

    returns
    -------
        2D ndarray (YX)

    """
    return img

def simple_sub(img, sub_img, scale=1.0):
    """
    Ffrom each pixel in *img*, subtract the corresponding
    pixel in *sub_img* multiplied by *scale*. Set all 
    negative values to zero.

    args
    ----
        img     :   2D ndarray (YX)
        sub_img :   2D ndarray (YX)
        scale   :   float

    returns
    -------
        2D ndarray (YX)

    """
    return img - scale * sub_img 

def sub_min(img, min_img, scale=1.0):
    """
    Wrapper for simple_sub() that uses the pixelwise minimum.

    args
    ----
        img     :   2D ndarray (YX)
        min_img :   2D ndarray (YX)
        scale   :   float

    returns
    -------
        2D ndarray (YX)

    """
    return simple_sub(img, min_img, scale=scale)

def sub_median(img, median_img, scale=1.0):
    """
    Wrapper for simple_sub() that uses the pixelwise median.

    args
    ----
        img        :   2D ndarray (YX)
        median_img :   2D ndarray (YX)
        scale      :   float

    returns
    -------
        2D ndarray (YX)

    """
    return simple_sub(img, median_img, scale=scale)   

def sub_mean(img, mean_img, scale=1.0):
    """
    Wrapper for simple_sub() that uses the pixelwise mean.

    args
    ----
        img        :   2D ndarray (YX)
        mean_img   :   2D ndarray (YX)
        scale      :   float

    returns
    -------
        2D ndarray (YX)

    """
    return simple_sub(img, mean_img, scale=scale)   

def sub_gauss_filt_min(img, gauss_filt_min_img, k=5.0, scale=1.0):
    """
    Subtract a Gaussian-filtered minimum image from this
    frame.

    args
    ----
        img         :   2D ndarray (YX)
        min_img     :   2D ndarray (YX)
        k           :   float, kernel sigma
        scale       :   float

    returns
    -------
        2D ndarray

    """
    return simple_sub(img, gauss_filt_min_img, scale=scale)

def sub_gauss_filt_median(img, gauss_filt_median_img, k=5.0, scale=1.0):
    """
    Subtract a Gaussian-filtered median image from this
    frame.

    args
    ----
        img         :   2D ndarray (YX)
        median_img  :   2D ndarray (YX)
        k           :   float, kernel sigma
        scale       :   float

    returns
    -------
        2D ndarray

    """
    return simple_sub(img, gauss_filt_median_img, scale=scale) 

def sub_gauss_filt_mean(img, gauss_filt_mean_img, k=5.0, scale=1.0):
    """
    Subtract a Gaussian-filtered mean image from this
    frame.

    args
    ----
        img         :   2D ndarray (YX)
        mean_img    :   2D ndarray (YX)
        k           :   float, kernel sigma
        scale       :   float

    returns
    -------
        2D ndarray

    """
    return simple_sub(img, gauss_filt_mean_img, scale=scale) 

# Factors to cache for each filtering method,
# removing redundant computations
FILTER_CACHES = {
    'identity': None,
    'sub_min': 'min_img',
    'sub_median': 'median_img',
    'sub_mean': 'mean_img',
    'sub_gauss_filt_min': 'gauss_filt_min_img',
    'sub_gauss_filt_median': 'gauss_filt_median_img',
    'sub_gauss_filt_mean': 'gauss_filt_mean_img',
}

# The base images for derived images
ORIGIN_IMAGES = {
    'gauss_filt_min_img': 'min_img',
    'gauss_filt_median_img': 'median_img',
    'gauss_filt_mean_img': 'mean_img',
}

# All available filtering methods
FILTER_METHODS = {
    'identity': identity,
    'sub_min': sub_min,
    'sub_median': sub_median,
    'sub_mean': sub_mean,
    'sub_gauss_filt_min': sub_gauss_filt_min,
    'sub_gauss_filt_median': sub_gauss_filt_median,
    'sub_gauss_filt_mean': sub_gauss_filt_mean,
}



#!/usr/bin/env python
"""
measureGain.py -- measure camera gain and BG with a simple
linear gain model

"""
# Numeric
import numpy as np 

# Fit gain model with LS
from scipy.optimize import curve_fit 

# Image file reader
from .read import ImageReader 

# Plotting utilities
from .plot import plot_pixel_mean_variance

def measure_camera_gain(*nd2_files, start_frame=100, plot=True):
    """
    Measure the camera gain and offset from a set of background movies.
    These can be used as the "camera_gain" and "camera_bg" arguments
    for localization settings, allowing the user to retrieve the PSF
    intensities in terms of photons rather than grayvalues.

    Each file in *nd2_files* should be a movie of an unlabeled, defocused,
    empty coverslip. Nothing should be movie, blinking, or exhibiting anything
    other than the ordinary camera noise. The movie should be acquired
    with exactly the same camera/stroboscopic settings as the SPT movies.

    Ideally, each movie in *nd2_files* should use a different level of 
    illumination, which facilitates more accurate measurement of the 
    gain.

    The method uses a pure Poisson noise model that is most applicable
    to EMCCD cameras.

    args
    ----
        nd2_files       :   variable number of str, background movies
        start_frame     :   int, the first frame in each movie to 
                            consider. This prevents photobleaching in
                            the early frames from influencing the 
                            result.

    returns
    -------
        dict, floats keyed to "camera_gain" and "camera_bg"

    """
    means, variances = [], []
    for i, nd2_file in enumerate(nd2_files):

        # Determine whether there are enough frames in this movie
        reader = ImageReader(nd2_file)
        if reader.n_frames <= start_frame:
            print("Warning: file {} only has {} frames".format(nd2_file, reader.n_frames))
            reader.close()
            continue 

        # Accumulate pixel means and variances for this movie
        I = np.zeros((reader.height, reader.width), dtype=np.float64)
        I2 = np.zeros((reader.height, reader.width), dtype=np.float64)
        n_frames = reader.n_frames - start_frame 
        for frame_idx in range(start_frame, reader.n_frames):
            frame = reader.get_frame(frame_idx)
            I += frame/n_frames 
            I2 += (frame**2)/n_frames 

        reader.close()

        # Pixel variance 
        I2 -= (I**2)

        # Append to the global list of means / variances
        means.append(I.ravel().copy())
        variances.append(I2.ravel().copy())

    # Flatten
    means_flat = np.concatenate(means)
    variances_flat = np.concatenate(variances)

    def linear_model(pixel_mean, gain, bg):
        """
        Given pixel mean, return pixel variance given a particular
        *gain* and *bg*.

        """
        return gain * (pixel_mean - bg)

    # Fit observed means and variances to a linear gain model
    popt, pcov = curve_fit(linear_model, means_flat, variances_flat, 
        bounds=(np.array([0, 0]), np.array([np.inf, np.inf])),
        p0=np.array([100.0, 400.0]))

    # Show the result to the user
    if plot:
        plot_pixel_mean_variance(means, variances, origin_files=nd2_files,
            model_gain=popt[0], model_bg=popt[1])

    return dict((
        ("camera_gain", popt[0]),
        ("camera_bg", popt[1])
    ))



from __future__ import annotations
from typing import Any
import napari
from napari.layers import Image
from qtpy import QtCore
from qtpy.QtWidgets import (
    QDialog,
    QWidget,
    QLabel,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QSplitter,
    QGroupBox,
    QFileDialog,
    QScrollArea,
    QTabWidget,
)

from magicgui.widgets import FileEdit
from functools import partial


from napari.qt import thread_worker
from magicgui.widgets import (
    create_widget,
    Container,
    ComboBox,
    FloatSpinBox,
    FloatSlider,
)
import enum
import dask.array as da
import dask_image
import pandas as pd
import logging
import os
from ..data_structure.acquisition import Acquisition
from .tif_pipeline import TifPipeline, ProcessingStep
from .edit_pipeline_window import PipelineEditor


class handled_types(enum.Enum):
    image = "Image"
    points = "Points"
    tracks = "Tracks"
    detections = "Detections"
    image_locs_and_pixel_size = "Image_locs_and_pixel_size"


class ProcessingStepWidget(Container):
    def __init__(self, step: ProcessingStep):
        self.widgets_dict = {}
        for param, value in step.__dict__.items():
            if param[0] == "_":
                continue
            logging.debug("Adding widget for %s" % param)
            w_type = (
                step.widget_types[param]
                if param in step.widget_types
                else None
            )
            w_options = (
                step.widget_options[param]
                if param in step.widget_options
                else {}
            )
            widget = create_widget(
                value=value,
                name=param,
                options=w_options,
                widget_type=w_type,
            )
            self.widgets_dict[param] = widget

            widget.changed.connect(partial(step.update_param, param))
        super().__init__(widgets=[w for _, w in self.widgets_dict.items()])


class TifPipelineWidget(QWidget):
    def __init__(
        self, tif_pipeline: TifPipeline, napari_viewer: napari.Viewer
    ):
        super().__init__()

        self.pixel_size = 0.093
        self.delta_t = 0.03
        self.to_export = None
        self.viewer = napari_viewer
        self.tp = tif_pipeline
        self._init_layers_dict()
        self.setup_ui()

    def activate_buttons(self, last_clicked: int = 0):
        for button_index, button in self._buttons.items():
            enable = button_index <= last_clicked + 1
            button.setEnabled(enable)

    def add_button(self, button: QPushButton):
        index = len(self._buttons)
        self._buttons[index] = button
        return index

    def drop_layers_after(self, first_idx_to_remove: int):
        original_layers = dict(self._layers)
        for idx, existing_layer in original_layers.items():
            if idx >= first_idx_to_remove:
                self.viewer.layers.remove(existing_layer)
                if idx in self._layers:
                    self._layers.pop(idx)
                logging.debug("Removed layer %d" % idx)

    def set_input_image(self, layer: Image = None):

        # Remove subsequent layers
        self.drop_layers_after(1)
        self.to_export = None
        self.export_locs_button.setEnabled(False)

        if layer is not None:
            # Convert data to Dask array
            # Store it
            logging.debug("Set Input Image %s" % layer.name)
            layer.data = layer.data.astype(float)
            if not isinstance(layer.data, da.Array):
                layer.data = da.from_array(layer.data)

            self._layers[0] = layer
            self.rescale_image_layers()

    def rescale_image_layers(self):
        rescaled = False
        for key, layer in self._layers.items():
            if isinstance(layer, Image):
                layer.scale = (1, self.pixel_size, self.pixel_size)
                rescaled = True
        if rescaled:
            self.viewer.reset_view()

    def setup_ui(self):
        main_layout = QVBoxLayout()

        main_layout.addLayout(self.choose_pipeline_layout())
        main_layout.addLayout(self.input_layout())

        main_layout.addWidget(QLabel("Processing steps :"))
        self.pipeline_widget = self.get_pipeline_widget()
        main_layout.addWidget(self.pipeline_widget)

        # main_layout.addStretch(-1)

        main_layout.addLayout(self.export_layout())

        self.setLayout(main_layout)

        self.init_buttons()
        self.viewer.layers.events.removed.connect(
            self._possibly_reset_pipeline
        )
        self.viewer.layers.events.inserted.connect(
            self._possibly_activate_button
        )

    def _init_layers_dict(self):
        self._layers = {}
        for layer in self.viewer.layers:
            if isinstance(layer, Image):
                self._layers[0] = layer

    def _possibly_reset_pipeline(self, event):
        key_of_removed = None
        original_items = dict(self._layers)
        for key, value in original_items.items():
            if event.value == value:
                key_of_removed = key
                break
        if key_of_removed is None:
            return
        print("Removed item had key %s" % key)
        for key, value in original_items.items():
            if key >= key_of_removed:
                self._layers.pop(key)
        self.activate_buttons(last_clicked=key - 2)

    def _possibly_activate_button(self, event):
        if isinstance(event.value, Image):
            self.activate_buttons(-1)

    def choose_pipeline_layout(self):

        main_layout = QHBoxLayout()
        main_layout.addWidget(QLabel("Load pipeline :"))
        self.select_file_widget = FileEdit(
            mode="r",
            filter="*.yaml",
            label="Pipeline file",
            name="Pipeline file",
        )
        self.select_file_widget.changed.connect(self.load_pipeline)
        main_layout.addWidget(self.select_file_widget._widget._qwidget)

        editPipelineButton = QPushButton("Edit current pipeline")

        def openEditWindow():
            editDialog = PipelineEditor(self.tp)
            if editDialog.exec():
                new_pipeline_widget = self.get_pipeline_widget()
                self.layout().replaceWidget(
                    self.pipeline_widget, new_pipeline_widget
                )
                self.pipeline_widget = new_pipeline_widget

        editPipelineButton.clicked.connect(openEditWindow)
        main_layout.addWidget(editPipelineButton)
        return main_layout

    def input_layout(self):
        main_layout = QVBoxLayout()
        input_widget: ComboBox = create_widget(
            annotation=Image,
            name="input_image",
            label="Input image : ",
            options={
                "tooltip": "Raw image on which the pipeline is run",
            },
        )
        input_widget.changed.connect(self.set_input_image)
        self.viewer.layers.events.inserted.connect(input_widget.reset_choices)
        self.viewer.layers.events.changed.connect(input_widget.reset_choices)
        self.viewer.layers.events.removed.connect(input_widget.reset_choices)
        self.viewer.layers.events.reordered.connect(input_widget.reset_choices)
        self.viewer.events.status.connect(input_widget.reset_choices)
        self.input_widget = input_widget

        delta_t_widget = FloatSpinBox(
            name="delta_t",
            label="delta t (s)",
            value=self.delta_t,
            step=0.005,
        )

        def set_delta_t(x: float):
            self.delta_t = x

        delta_t_widget.changed.connect(set_delta_t)

        pixel_size_widget = FloatSpinBox(
            name="pixel_size",
            label="pixel size (um)",
            value=self.pixel_size,
            step=0.001,
        )

        def set_pixel_size(x: float):
            self.pixel_size = x
            self.rescale_image_layers()

        pixel_size_widget.changed.connect(set_pixel_size)

        container = Container(
            widgets=[input_widget, delta_t_widget, pixel_size_widget]
        )
        main_layout.addWidget(container._widget._qwidget)

        main_box = QGroupBox("Input : ")
        main_box.setLayout(main_layout)
        main_layout = QVBoxLayout()
        main_layout.addWidget(main_box)

        return main_layout

    def load_pipeline(self, file_path):
        try:
            tp = TifPipeline.from_yaml(file_path)
        except BaseException as e:
            logging.error(e)
            logging.error(
                "Could not load pipeline from file %s, check the format"
                % file_path
            )
            return
        self.tp = tp
        self.pipeline_widget.setWidget(self.get_pipeline_widget())
        self.setWindowTitle("Palmari | %s" % tp.name)
        self.init_buttons()

    def init_buttons(self):
        self.activate_buttons(last_clicked=-1 if 0 in self._layers else -2)

    def get_pipeline_widget(self):

        self._buttons = {}

        pipeline_widget = QWidget()
        pipeline_layout = QVBoxLayout()

        preprocessing_widget = self.preprocessing_widget()
        if preprocessing_widget is not None:
            pipeline_layout.addWidget(preprocessing_widget)

        pipeline_layout.addWidget(self.detection_widget())
        pipeline_layout.addWidget(self.localization_widget())

        post_processing_widget = self.locs_processing_widget()
        if post_processing_widget is not None:
            pipeline_layout.addWidget(post_processing_widget)

        pipeline_layout.addWidget(self.tracking_widget())
        pipeline_widget.setLayout(pipeline_layout)
        pipeline_widget.setContentsMargins(0, 0, 10, 0)

        pipeline_container = QScrollArea()
        pipeline_container.setWidget(pipeline_widget)
        pipeline_container.setWidgetResizable(False)
        pipeline_widget.setFixedHeight(pipeline_widget.sizeHint().height())
        pipeline_container.setWidgetResizable(True)

        return pipeline_container

    def export_layout(self):
        main_layout = QVBoxLayout()
        self.export_locs_button = QPushButton(text="Save locs and tracks")
        export_pipeline_button = QPushButton(text="Save pipeline")
        main_layout.addWidget(self.export_locs_button)
        main_layout.addWidget(export_pipeline_button)

        self.export_locs_button.clicked.connect(self.export_locs)
        export_pipeline_button.clicked.connect(self.export_pipeline)

        return main_layout

    def export_locs(self):
        assert self.to_export is not None
        fileName, _ = QFileDialog.getSaveFileName(
            self,
            caption="Export localizations and tracks",
        )
        logging.debug(fileName)
        if fileName is not None and len(fileName) > 2:
            self.to_export.to_csv(fileName, index=True)

    def export_pipeline(self):
        fileName, _ = QFileDialog.getSaveFileName(
            self,
            directory=self.tp.last_storage_path,
            caption="Save pipeline parameters",
        )
        logging.debug(fileName)

        if fileName is not None and len(fileName) > 2:
            if ".yaml" not in fileName:
                fileName = "%s.yaml" % fileName
            self.tp.to_yaml(fileName)

    def preprocessing_widget(self):

        if len(self.tp.movie_preprocessors) == 0:
            return None

        main_widget = QGroupBox(title="Image pre-processing")
        main_layout = QVBoxLayout()
        for i, step in enumerate(self.tp.movie_preprocessors):
            step_layout = self.setup_layout_for_step(
                step,
                input_type=handled_types.image,
                output_type=handled_types.image,
            )
            main_layout.addLayout(step_layout)
            if i < len(self.tp.movie_preprocessors) - 1:
                main_layout.addWidget(QSplitter())
        main_widget.setLayout(main_layout)

        return main_widget

    def detection_widget(self):

        main_widget = QGroupBox(title="Detection : %s" % self.tp.detector.name)
        main_layout = QVBoxLayout()
        main_layout.addLayout(
            self.setup_layout_for_step(
                self.tp.detector,
                input_type=handled_types.image,
                output_type=handled_types.detections,
                skip_title=True,
            )
        )
        main_widget.setLayout(main_layout)
        return main_widget

    def localization_widget(self):

        main_widget = QGroupBox(
            title="Localization : %s" % self.tp.localizer.name
        )
        main_layout = QVBoxLayout()
        main_layout.addLayout(
            self.setup_layout_for_step(
                self.tp.localizer,
                input_type=handled_types.detections,
                output_type=handled_types.points,
                skip_title=True,
            )
        )
        main_widget.setLayout(main_layout)
        return main_widget

    def locs_processing_widget(self):

        if len(self.tp.loc_processors) == 0:
            return None

        main_widget = QGroupBox(title="Post-processing of localizations")
        main_layout = QVBoxLayout()
        for i, step in enumerate(self.tp.loc_processors):
            step_layout = self.setup_layout_for_step(
                step,
                input_type=handled_types.image_locs_and_pixel_size,
                output_type=handled_types.points,
            )
            main_layout.addLayout(step_layout)
            if i < len(self.tp.loc_processors) - 1:
                main_layout.addWidget(QSplitter())
        main_widget.setLayout(main_layout)

        return main_widget

    def tracking_widget(self):

        main_widget = QGroupBox(title="Tracking : %s" % self.tp.tracker.name)
        main_layout = QVBoxLayout()
        main_layout.addLayout(
            self.setup_layout_for_step(
                self.tp.tracker,
                input_type=handled_types.points,
                output_type=handled_types.tracks,
                skip_title=True,
                is_final_step=True,
            )
        )
        main_widget.setLayout(main_layout)
        return main_widget

    def setup_layout_for_step(
        self,
        step: ProcessingStep,
        input_type: handled_types,
        output_type: handled_types,
        skip_title: bool = False,
        is_final_step: bool = False,
    ) -> QVBoxLayout:
        step_layout = QVBoxLayout()
        if not skip_title:
            step_layout.addWidget(QLabel(step.name))
        step_container = ProcessingStepWidget(step)
        if len(step_container.widgets_dict) > 0:
            step_layout.addWidget(step_container._widget._qwidget)
        step_run_btn = QPushButton(step.action_name)
        btn_index = self.add_button(step_run_btn)
        step_layout.addWidget(step_run_btn)
        input_layer_idx = btn_index
        output_layer_idx = btn_index + 1

        def add_layer(data):
            logging.debug(
                "Adding layer %d -> %s" % (input_layer_idx, output_layer_idx)
            )
            if step.is_localizer:
                data["t"] = (
                    data["frame"] - data["frame"].min()
                ) * self.delta_t
                data[["x", "y"]] *= self.pixel_size

            if step.is_detector:
                data[["x", "y"]] = data[["y", "x"]]

            self.drop_layers_after(output_layer_idx)
            if data.shape[0] > 0:
                self.add_layer_of_type(
                    data=data,
                    data_type=output_type,
                    layer_idx=output_layer_idx,
                    step_name=step.name,
                )
            self.setEnabled(True)
            if data.shape[0] > 0:
                self.activate_buttons(last_clicked=btn_index)
            else:
                self.activate_buttons(last_clicked=btn_index - 1)
            if is_final_step:
                self.to_export = data
            else:
                self.to_export = None
            self.export_locs_button.setEnabled(is_final_step)

        @thread_worker(connect={"returned": add_layer})
        def run_step(checked: bool = True):
            logging.debug(self._layers)
            assert input_layer_idx in self._layers
            self.setEnabled(False)
            if input_type == handled_types.image:
                input_data = (self._layers[input_layer_idx].data,)
            elif input_type == handled_types.points:
                input_data = (self._layers[input_layer_idx].result,)
            elif input_type == handled_types.tracks:
                input_data = (self._layers[input_layer_idx].result,)
            elif input_type == handled_types.detections:
                detections = pd.DataFrame(
                    data=self._layers[input_layer_idx].data.copy(),
                    columns=["frame", "x", "y"],
                )
                detections[["x", "y"]] /= self.pixel_size
                # On inverse x et y volontairement
                detections[["y", "x"]] = detections[["x", "y"]].astype(int)
                input_data = (
                    self._layers[input_layer_idx - 1].data,
                    detections,
                )
            elif input_type == handled_types.image_locs_and_pixel_size:
                # detections = pd.DataFrame(
                #    data=self._layers[input_layer_idx].data.copy(),
                #    columns=["frame", "x", "y"],
                # )
                # detections[["x", "y"]] /= self.pixel_size
                # On inverse x et y volontairement
                # detections[["y", "x"]] = detections[["x", "y"]].astype(int)
                input_data = (
                    self._layers[0].data,
                    self._layers[input_layer_idx].result,
                    self.pixel_size,
                )
            return step.process(*input_data)

        step_run_btn.clicked.connect(run_step)

        return step_layout

    def add_layer_of_type(
        self,
        data: Any,
        data_type: handled_types,
        layer_idx: int,
        step_name: str,
    ):
        if data is None:
            logging.debug("Nothing returned by %s" % step_name)

        if data_type == handled_types.image:
            self.add_image_layer(
                returned_image=data, layer_idx=layer_idx, step_name=step_name
            )
        elif data_type == handled_types.tracks:
            self.add_tracks_layer(
                tracks=data, layer_idx=layer_idx, step_name=step_name
            )
        elif data_type == handled_types.points:
            self.add_points_layer(
                points=data,
                layer_idx=layer_idx,
                step_name=step_name,
            )
        elif data_type == handled_types.detections:
            self.add_points_layer(
                points=data, layer_idx=layer_idx, step_name=step_name
            )

    def add_image_layer(self, returned_image, layer_idx: int, step_name: str):

        if layer_idx not in self._layers:
            new_layer = self.viewer.add_image(
                data=returned_image,
                name="%s : %s" % (self._layers[0].name, step_name),
                scale=(1, self.pixel_size, self.pixel_size),
            )
            self._layers[layer_idx] = new_layer
        else:
            self._layers[layer_idx].data = returned_image

    def add_points_layer(
        self,
        points,
        layer_idx: int,
        step_name: str,
    ):

        if layer_idx not in self._layers:
            new_layer = self.viewer.add_points(
                points[["frame", "x", "y"]].values,
                properties=points[
                    [
                        c
                        for c in points.columns
                        if (c not in ["x", "y"])
                        and (points[c].dtype.kind in "uifb")
                    ]
                ].to_dict(),
                symbol="disc",
                size=0.25,
                edge_width=0.1,
                face_color="transparent" if (points.shape[0] > 0) else None,
                edge_color="frame" if (points.shape[0] > 0) else None,
                edge_colormap="rainbow",
                blending="translucent_no_depth",
                name="%s : %s" % (self._layers[0].name, step_name),
            )
            new_layer.editable = False
            self._layers[layer_idx] = new_layer
        else:
            self._layers[layer_idx].data = points

        self._layers[layer_idx].result = points

    def add_tracks_layer(self, tracks, layer_idx: int, step_name: str):

        if layer_idx not in self._layers:
            new_layer = self.viewer.add_tracks(
                data=tracks[["n", "frame", "x", "y"]],
                tail_width=1,
                tail_length=10,
                head_length=0,
                properties=pd.merge(
                    tracks[["n"]],
                    tracks.groupby("n")
                    .agg({"x": "count", "frame": "min"})
                    .rename(columns={"x": "length", "frame": "start_frame"}),
                    how="left",
                    left_on="n",
                    right_index=True,
                ),
                name="%s : %s" % (self._layers[0].name, step_name),
            )
            self._layers[layer_idx] = new_layer
        else:
            self._layers[layer_idx].tracks = tracks

        self._layers[layer_idx].result = tracks

    @classmethod
    def view_pipeline(
        cls,
        tif_pipeline: TifPipeline,
        acq: Acquisition = None,
        tif_file: str = None,
    ):
        napari.Viewer()
        viewer = napari.current_viewer()
        widget = cls(tif_pipeline, viewer)

        if acq is not None:
            widget.pixel_size = acq.experiment.pixel_size
            widget.delta_t = acq.experiment.DT
            viewer.add_image(
                data=acq.image,
                name=acq.tif_file.split(os.path.sep)[-1],
                scale=(1.0, widget.pixel_size, widget.pixel_size),
            )
        elif tif_file is not None:
            viewer.add_image(
                data=dask_image.imread.imread(
                    tif_file,
                    nframes=300,
                )
            )

        viewer.window.add_dock_widget(
            widget=widget,
            name="PALM pipeline : %s" % tif_pipeline.name,
            area="right",
        )
        widget.rescale_image_layers()
